# -*- coding: utf-8 -*-
"""
This submodule provides common metrics used in 3D orientation representations,
focusing on the validation and comparison of rotation matrices and quaternions.

They include guard clauses to raise appropriate errors if the inputs do not
meet the required criteria, ensuring robust and error-free computations.

Additionally, the submodule implements a euclidean error estimation function,
and functions to calculate the RMS error between two matrices or between an
array of matrices, either element-wise or as a whole.

"""

import numpy as np
from ..common.dcm import DCM
from typing import Union

def _rotations_guard_clauses(R1: Union[list, np.ndarray], R2: Union[list, np.ndarray]) -> None:
    """
    Checks validity of rotation matrices.

    Raises
    ------
    TypeError
        If the rotation matrices are not valid types.
    ValueError
        If the rotation matrices are not valid shapes.

    """
    for label, rotation_matrix in zip(['R1', 'R2'], [R1, R2]):
        if not isinstance(rotation_matrix, (list, np.ndarray)):
            raise TypeError(f"{label} must be an array. Got {type(rotation_matrix)}")
    r1, r2 = np.copy(R1), np.copy(R2)
    for rotation_matrix in [r1, r2]:
        if rotation_matrix.shape[-2:] != (3, 3):
            raise ValueError(f"Rotation matrices must be of shape (N, 3, 3) or (3, 3). Got {rotation_matrix.shape}.")
    r1_shape, r2_shape = r1.shape, r2.shape
    if r1_shape != r2_shape:
        raise ValueError(f"Cannot compare R1 of shape {r1_shape} and R2 of shape {r2_shape}.")

def _quaternions_guard_clauses(q1: Union[list, np.ndarray], q2: Union[list, np.ndarray]) -> None:
    """
    Checks validity of quaternions.

    Raises
    ------
    TypeError
        If the quaternions are not valid types.
    ValueError
        If the quaternions are not valid shapes.

    """
    for label, quaternion in zip(['q1', 'q2'], [q1, q2]):
        if not isinstance(quaternion, (list, np.ndarray)):
            raise TypeError(f"{label} must be an array. Got {type(quaternion)}")
    q1, q2 = np.copy(q1), np.copy(q2)
    for quaternion in [q1, q2]:
        if quaternion.shape[-1] != 4:
            raise ValueError(f"Quaternions must be of shape (N, 4) or (4,). Got {quaternion.shape}.")
    if q1.shape != q2.shape:
        raise ValueError(f"Cannot compare q1 of shape {q1.shape} and q2 of shape {q2.shape}")

def euclidean(x: np.ndarray, y: np.ndarray) -> float:
    """
    Euclidean distance between Euler angles as described in :cite:p:`huynh2009`.

    Given two sets of Euler angles :math:`\\mathbf{x}` and :math:`\\mathbf{y}`,
    the Euclidean distance between them is computed as:

    .. math::
        \\Phi(\\mathbf{x}, \\mathbf{y}) = \\sqrt{d(x_0, y_0)^2 + d(x_2, y_1)^2 + d(x_2, y_2)^2}

    where the function :math:`d(\\mathbf{x}, \\mathbf{y})` is the normalized
    difference between the two angles so that :math:`0 \\leq d(\\cdot, \\cdot) \\leq \\pi`.

    .. math::

        d(\\mathbf{x}, \\mathbf{y}) = \\min \\{ |\\mathbf{x}-\\mathbf{y}|, 2\\pi - |\\mathbf{x}-\\mathbf{y}| \\}

    .. warning::

        This metric is only valid for Euler angles in the range :math:`[0, 2\\pi]`
        and is not in SO(3), since it depends on a representation that is not
        unique.

    This metric gives values in the range [0, :math:`\\pi\\sqrt{3}`]

    Parameters
    ----------
    x : numpy.ndarray
        M-by-N array to compare. Usually a reference array.
    y : numpy.array
        M-by-N array to compare.
    mode : str
        Mode of distance computation.

    Return
    ------
    d : float
        Distance or difference between arrays.

    Examples
    --------
    >>> import numpy as np
    >>> from ahrs.utils.metrics import euclidean
    >>> num_samples = 5
    >>> angles = np.random.uniform(low=-180.0, high=180.0, size=(num_samples, 3))
    >>> noisy = angles + np.random.randn(num_samples, 3)
    >>> euclidean(angles, noisy)
    2.585672169476804
    >>> euclidean(angles, noisy, axis=0)
    array([1.36319772, 1.78554071, 1.28032688])
    >>> euclidean(angles, noisy, axis=1)     # distance per sample
    array([0.88956871, 1.19727356, 1.5243858 , 0.68765523, 1.29007067])

    """
    x, y = np.copy(x), np.copy(y)
    if x.shape != y.shape:
        raise ValueError("Both arrays must have the same shape.")
    a_b = abs(x - y)
    d = np.minimum(a_b, 2*np.pi - a_b)
    if x.ndim == 1:
        return np.linalg.norm(d)
    return np.linalg.norm(d, axis=1)

def chordal(R1: np.ndarray, R2: np.ndarray) -> Union[float, np.ndarray]:
    """
    The chordal distance between two rotations :math:`\\mathbf{R}_1` and
    :math:`\\mathbf{R}_2` in SO(3) is the Euclidean distance between them in
    the embedding space :math:`\\mathbb{R}^{3\\times 3}=\\mathbb{R}^9`
    :cite:p:`hartley2013`:

    .. math::
        d(\\mathbf{R}_1, \\mathbf{R}_2) = \\|\\mathbf{R}_1-\\mathbf{R}_2\\|_F

    where :math:`\\|\\mathbf{X}\\|_F` represents the Frobenius norm of the
    matrix :math:`\\mathbf{X}`.

    The error lies within: [0, :math:`2\\sqrt{3}`]

    Parameters
    ----------
    R1 : numpy.ndarray
        3-by-3 rotation matrix.
    R2 : numpy.ndarray
        3-by-3 rotation matrix.

    Returns
    -------
    d : float or numpy.ndarray
        Chordal distance between matrices.

    Raises
    ------
    ValueError
        If the rotation matrices are not of the same shape.
    TypeError
        If the rotation matrices are not of type list or numpy.ndarray.

    Examples
    --------
    >>> import ahrs
    >>> R1 = ahrs.DCM(rpy=[0.0, 0.0, 0.0])
    >>> R2 = ahrs.DCM(rpy=[90.0, 90.0, 90.0])
    >>> ahrs.utils.chordal(R1, R2)
    2.0
    >>> R1 = ahrs.DCM(rpy=[10.0, -20.0, 30.0])
    >>> R2 = ahrs.DCM(rpy=[-10.0, 20.0, -30.0])
    >>> ahrs.utils.chordal(R1, R2)
    1.6916338074634352

    """
    _rotations_guard_clauses(R1, R2)
    R1, R2 = np.copy(R1), np.copy(R2)
    if R1.ndim < 3:
        return np.linalg.norm(R1-R2, 'fro')
    return np.array([np.linalg.norm(r1-r2, 'fro') for r1, r2 in zip(R1, R2)])

def identity_deviation(R1: np.ndarray, R2: np.ndarray) -> float:
    """
    The deviation from the identity matrix between two rotations :math:`\\mathbf{R}_1`
    and :math:`\\mathbf{R}_2` in SO(3) is defined in :cite:p:`huynh2009` as:

    .. math::
        d(\\mathbf{R}_1, \\mathbf{R}_2) = \\|\\mathbf{I}-\\mathbf{R}_1\\mathbf{R}_2^T\\|_F

    where :math:`\\|\\mathbf{X}\\|_F` represents the Frobenius norm of the
    matrix :math:`\\mathbf{X}`.

    If the given matrices are close to each other, the resulting deviation will
    be close to zero.

    The error lies within: [0, :math:`2\\sqrt{3}`]

    Parameters
    ----------
    R1 : numpy.ndarray
        3-by-3 rotation matrix.
    R2 : numpy.ndarray
        3-by-3 rotation matrix.

    Returns
    -------
    d : float
        Deviation from identity matrix.

    Raises
    ------
    ValueError
        If the rotation matrices are not of the same shape.
    TypeError
        If the rotation matrices are not of type list or numpy.ndarray.

    Examples
    --------
    >>> import ahrs
    >>> R1 = ahrs.DCM(rpy=[0.0, 0.0, 0.0])
    >>> R2 = ahrs.DCM(rpy=[90.0, 90.0, 90.0])
    >>> ahrs.utils.identity_deviation(R1, R2)
    2.0
    >>> R1 = ahrs.DCM(rpy=[10.0, -20.0, 30.0])
    >>> R2 = ahrs.DCM(rpy=[-10.0, 20.0, -30.0])
    >>> ahrs.utils.identity_deviation(R1, R2)
    1.6916338074634352

    """
    _rotations_guard_clauses(R1, R2)
    R1, R2 = np.copy(R1), np.copy(R2)
    return np.linalg.norm(np.eye(3)-R1@R2.T, 'fro')

def angular_distance(R1: np.ndarray, R2: np.ndarray) -> float:
    """
    The angular distance between two rotations :math:`\\mathbf{R}_1` and
    :math:`\\mathbf{R}_2` in SO(3), as defined in :cite:p:`hartley2013`:

    .. math::
        d(\\mathbf{R}_1, \\mathbf{R}_2) = \\|\\log(\\mathbf{R}_1\\mathbf{R}_2^T)\\|

    where :math:`\\|\\mathbf{x}\\|` represents the usual euclidean norm of the
    vector :math:`\\mathbf{x}`.

    Parameters
    ----------
    R1 : numpy.ndarray
        3-by-3 rotation matrix.
    R2 : numpy.ndarray
        3-by-3 rotation matrix.

    Returns
    -------
    d : float
        Angular distance between rotation matrices

    Examples
    --------
    >>> import ahrs
    >>> R1 = ahrs.DCM(rpy=[0.0, 0.0, 0.0])
    >>> R2 = ahrs.DCM(rpy=[90.0, 90.0, 90.0])
    >>> ahrs.utils.angular_distance(R1, R2)
    1.5707963267948966
    >>> R1 = ahrs.DCM(rpy=[10.0, -20.0, 30.0])
    >>> R2 = ahrs.DCM(rpy=[-10.0, 20.0, -30.0])
    >>> ahrs.utils.angular_distance(R1, R2)
    1.282213683073497

    """
    _rotations_guard_clauses(R1, R2)
    R1, R2 = np.copy(R1), np.copy(R2)
    R1R2T = DCM(R1@R2.T)
    return np.linalg.norm(R1R2T.log)

def qdist(q1: np.ndarray, q2: np.ndarray) -> float:
    """
    Euclidean distance between two unit quaternions as defined in
    :cite:p:`huynh2009` and :cite:p:`hartley2013`:

    .. math::
        d(\\mathbf{q}_1, \\mathbf{q}_2) = \\mathrm{min} \\{ \\|\\mathbf{q}_1-\\mathbf{q}_2\\|, \\|\\mathbf{q}_1+\\mathbf{q}_2\\|\\}

    Both quaternions will be normalized before they are compared.

    The error lies within [0, :math:`\\sqrt{2}`], where :math:`d` is equal to
    zero if the quaternions represent the same attitude, and :math:`\\sqrt{2}`
    if they are completely divergent.

    Parameters
    ----------
    q1 : numpy.ndarray
        First quaternion, or set of quaternions, to compare.
    q2 : numpy.ndarray
        Second quaternion, or set of quaternions, to compare.

    Returns
    -------
    d : float
        Euclidean distance between given unit quaternions.

    Raises
    ------
    ValueError
        If given quaternions do not have the same shape.
    TypeError
        If given quaternions are not of type list or numpy.ndarray.

    Examples
    --------
    >>> q1 = ahrs.Quaternion(random=True)
    >>> q1.view()
    Quaternion([ 0.94185064,  0.04451339, -0.00622856,  0.33301221])
    >>> q2 = ahrs.Quaternion(random=True)
    >>> q2.view()
    Quaternion([-0.51041283, -0.38336653,  0.76929238, -0.0264211 ])
    >>> ahrs.utils.qdist(q1, q2)
    0.9885466801358284
    >>> ahrs.utils.qdist(q1, -q1)
    0.0
    """
    _quaternions_guard_clauses(q1, q2)
    q1, q2 = np.copy(q1), np.copy(q2)
    if q1.ndim == 1:
        q1 /= np.linalg.norm(q1)
        q2 /= np.linalg.norm(q2)
        if np.allclose(q1, q2) or np.allclose(-q1, q2):
            return 0.0
        return min(np.linalg.norm(q1-q2), np.linalg.norm(q1+q2))
    q1 /= np.linalg.norm(q1, axis=1)[:, None]
    q2 /= np.linalg.norm(q2, axis=1)[:, None]
    return np.r_[[np.linalg.norm(q1-q2, axis=1)], [np.linalg.norm(q1+q2, axis=1)]].min(axis=0)

def qeip(q1: np.ndarray, q2: np.ndarray) -> float:
    """
    Euclidean distance of inner products between two unit quaternions
    :math:`\\mathbf{q}_1` and :mathbf{q}_2` as defined in :cite:p:`huynh2009`
    and :cite:p:`kuffner2004`:

    .. math::
        d(\\mathbf{q}_1, \\mathbf{q}_2) = 1 - |\\mathbf{q}_1\\cdot\\mathbf{q}_2|

    The error lies within [0, 1], where :math:`d` is equal to zero if the
    quaternions represent the same attitude, and to one if they are completely
    divergent.

    Parameters
    ----------
    q1 : numpy.ndarray
        First quaternion, or set of quaternions, to compare.
    q2 : numpy.ndarray
        Second quaternion, or set of quaternions, to compare.

    Returns
    -------
    d : float
        Euclidean distance of inner products between given unit quaternions.

    Raises
    ------
    ValueError
        If the quaternions are not of the same shape.
    TypeError
        If the quaternions are not of type list or numpy.ndarray.

    Examples
    --------
    >>> q1 = ahrs.Quaternion(random=True)
    >>> q1.view()
    Quaternion([ 0.94185064,  0.04451339, -0.00622856,  0.33301221])
    >>> q2 = ahrs.Quaternion(random=True)
    >>> q2.view()
    Quaternion([-0.51041283, -0.38336653,  0.76929238, -0.0264211 ])
    >>> ahrs.utils.qeip(q1, q2)
    0.48861226940378377
    >>> ahrs.utils.qeip(q1, -q1)
    0.0
    """
    _quaternions_guard_clauses(q1, q2)
    q1, q2 = np.copy(q1), np.copy(q2)
    if q1.ndim == 1:
        q1 /= np.linalg.norm(q1)
        q2 /= np.linalg.norm(q2)
        if np.allclose(q1, q2) or np.allclose(-q1, q2):
            return 0.0
        return 1.0-abs(q1@q2)
    q1 /= np.linalg.norm(q1, axis=1)[:, None]
    q2 /= np.linalg.norm(q2, axis=1)[:, None]
    return 1.0-abs(np.nansum(q1*q2, axis=1))

def qcip(q1: np.ndarray, q2: np.ndarray) -> float:
    """
    The cosine of inner products between two unit quaternions
    :math:`\\mathbf{q}_1` and :mathbf{q}_2` as defined in :cite:p:`huynh2009`:

    .. math::
        d(\\mathbf{q}_1, \\mathbf{q}_2) = \\arccos(|\\mathbf{q}_1\\cdot\\mathbf{q}_2|)

    The error lies within: [0, :math:`\\frac{\\pi}{2}`]

    Parameters
    ----------
    q1 : numpy.ndarray
        First quaternion, or set of quaternions, to compare.
    q2 : numpy.ndarray
        Second quaternion, or set of quaternions, to compare.

    Returns
    -------
    d : float
        Cosine of inner products of quaternions.

    Raises
    ------
    ValueError
        If the quaternions are not of the same shape.
    TypeError
        If the quaternions are not of type list or numpy.ndarray.

    Examples
    --------
    >>> q1 = ahrs.Quaternion(random=True)
    >>> q1.view()
    Quaternion([ 0.94185064,  0.04451339, -0.00622856,  0.33301221])
    >>> q2 = ahrs.Quaternion(random=True)
    >>> q2.view()
    Quaternion([-0.51041283, -0.38336653,  0.76929238, -0.0264211 ])
    >>> ahrs.utils.qcip(q1, q2)
    1.0339974504196667
    >>> ahrs.utils.qcip(q1, -q1)
    0.0
    """
    _quaternions_guard_clauses(q1, q2)
    q1, q2 = np.copy(q1), np.copy(q2)
    if q1.ndim == 1:
        q1 /= np.linalg.norm(q1)
        q2 /= np.linalg.norm(q2)
        if np.allclose(q1, q2) or np.allclose(-q1, q2):
            return 0.0
        return np.arccos(abs(q1@q2))
    q1 /= np.linalg.norm(q1, axis=1)[:, None]
    q2 /= np.linalg.norm(q2, axis=1)[:, None]
    return np.arccos(abs(np.nansum(q1*q2, axis=1)))

def qad(q1: np.ndarray, q2: np.ndarray) -> float:
    """
    The quaternion angular difference between two unit quaternions
    :math:`\\mathbf{q}_1` and :mathbf{q}_2` as defined in :cite:p:`thibaud2017`:

    .. math::

        d(\\mathbf{q}_1, \\mathbf{q}_2) = \\arccos(2\\langle\\mathbf{q}_1,\\mathbf{q}_2\\rangle^2-1)

    The error lies within: [0, :math:`\\pi`]

    Parameters
    ----------
    q1 : numpy.ndarray
        First quaternion, or set of quaternions, to compare.
    q2 : numpy.ndarray
        Second quaternion, or set of quaternions, to compare.

    Returns
    -------
    d : float
        Angle difference between given unit quaternions.

    Raises
    ------
    ValueError
        If the quaternions are not of the same shape.
    TypeError
        If the quaternions are not of type list or numpy.ndarray.

    Examples
    --------
    >>> q1 = ahrs.Quaternion(random=True)
    >>> q1.view()
    Quaternion([ 0.94185064,  0.04451339, -0.00622856,  0.33301221])
    >>> q2 = ahrs.Quaternion(random=True)
    >>> q2.view()
    Quaternion([-0.51041283, -0.38336653,  0.76929238, -0.0264211 ])
    >>> ahrs.utils.qad(q1, q2)
    2.0679949008393335
    >>> ahrs.utils.qad(q1, -q1)
    0.0

    """
    _quaternions_guard_clauses(q1, q2)
    q1, q2 = np.copy(q1), np.copy(q2)
    if q1.ndim == 1:
        q1 /= np.linalg.norm(q1)
        q2 /= np.linalg.norm(q2)
        if np.allclose(q1, q2) or np.allclose(-q1, q2):
            return 0.0
        return np.arccos(2.0*(q1@q2)**2-1.0)
    q1 /= np.linalg.norm(q1, axis=1)[:, None]
    q2 /= np.linalg.norm(q2, axis=1)[:, None]
    return np.arccos(np.clip(2.0*np.sum(q1*q2, axis=1)**2-1.0, -1.0, 1.0))

def rmse(x: np.ndarray, y: np.ndarray):
    """
    The `RMSE <https://en.wikipedia.org/wiki/Root_mean_square_deviation>`_ is
    computed as:

    .. math::

        d(\\mathbf{x}, \\mathbf{y}) = \\sqrt{\\frac{1}{N}\\sum_{i=1}^N (x_i-y_i)^2}

    where :math:`N` is the number of elements in :math:`\\mathbf{x}` and
    :math:`\\mathbf{y}`.

    If :math:`\\mathbf{x}` and :math:`\\mathbf{y}` are :math:`M \\times N`
    matrices, then the RMSE is computed for each row, yielding a vector of
    length :math:`M`.

    Parameters
    ----------
    x : numpy.ndarray
        First set of values to compare.
    y : numpy.ndarray
        Second set of values to compare.

    Returns
    -------
    d : float
        Root mean squared error between given values.
    """
    x, y = np.copy(x), np.copy(y)
    if x.ndim > 1:
        return np.sqrt(np.nanmean((x-y)**2, axis=1))
    return np.sqrt(np.nanmean((x-y)**2))

def rmse_matrices(A: np.ndarray, B: np.ndarray, element_wise: bool = False) -> np.ndarray:
    """
    The Root Mean Square Error (RMSE) between two arrays of matrices :math:`A`
    and :math:`B` is calculated as the square root of the mean of the squared
    differences between the elements of the two arrays.

    If the input arrays are 2-dimensional matrices, the RMSE is calculated as:

    .. math::
        RMSE = \\sqrt{\\frac{1}{MN}\\sum_{i=1}^{M}\\sum_{j=1}^{N}(A_{ij} - B_{ij})^2}

    If the input arrays are arrays of 2-dimensional matrices (3-dimensional
    array), the RMSE is calculated as:

    .. math::
        RMSE = \\sqrt{\\frac{1}{k}\\sum_{l=1}^{k}\\frac{1}{MN}\\sum_{i=1}^{M}\\sum_{j=1}^{N}(A_{ij}^{(l)} - B_{ij}^{(l)})^2}

    where :math:`k` is the number of matrices in the arrays.

    If the option ``element_wise`` is set to ``True``, the RMSE is calculated
    element-wise, and an M-by-N array of RMSEs is returned. The following calls
    are equivalent:

    .. code-block:: python

        rmse = rmse_matrices(A, B, element_wise=True)
        rmse = np.sqrt(np.nanmean((A-B)**2, axis=0))

    If the inputs are arrays of matrices (3-dimensional arrays), its call is
    also equivalent to:

    .. code-block:: python

        rmse = np.zeros_like(A[0])
        for i in range(A.shape[1]):
            for j in range(A.shape[2]):
                rmse[i, j] = np.sqrt(np.nanmean((A[:, i, j]-B[:, i, j])**2))

    If the inputs are 2-dimensional matrices, the following calls would return
    the same result:

    .. code-block:: python

        rmse_matrices(A, B)
        rmse_matrices(A, B, element_wise=False)
        rmse_matrices(A, B, element_wise=True)

    Parameters
    ----------
    A : numpy.ndarray
        First M-by-N matrix or array of k M-by-N matrices.
    B : numpy.ndarray
        Second M-by-N matrix or array of k M-by-N matrices.
    element_wise : bool, default: False
        If True, calculate RMSE element-wise, and return an M-by-N array of
        RMSEs.

    Returns
    -------
    rmse : float or np.ndarray
        Root Mean Square Error between the two matrices, or array of k RMSEs
        between the two arrays of matrices.

    Raises
    ------
    ValueError
        If the comparing arrays do not have the same shape.

    Examples
    --------
    .. code-block:: python

        >>> C = np.random.random((4, 3, 2))     # Array of four 3-by-2 matrices
        >>> C.view()
        array([[[0.2816407 , 0.30850589],
                [0.44618209, 0.33081522],
                [0.7994625 , 0.07377569]],

               [[0.35549399, 0.47050713],
                [0.94168683, 0.50388058],
                [0.70023837, 0.77216167]],

               [[0.79897129, 0.28555452],
                [0.892488  , 0.71476669],
                [0.19071524, 0.4123666 ]],

               [[0.86301978, 0.14686002],
                [0.98784823, 0.26129908],
                [0.46982206, 0.88037599]]])
        >>> D = np.random.random((4, 3, 2))     # Array of four 3-by-2 matrices
        >>> D.view()
        array([[[0.71560918, 0.34100321],
                [0.92518341, 0.50741267],
                [0.30730944, 0.19173378]],

               [[0.31846657, 0.08578454],
                [0.62643489, 0.84014104],
                [0.7111152 , 0.95428613]],

               [[0.8101591 , 0.9584096 ],
                [0.91118705, 0.71203119],
                [0.58217189, 0.45598271]],

               [[0.79837603, 0.09954558],
                [0.26532781, 0.55711476],
                [0.03909648, 0.10787888]]])
        >>> rmse_matrices(C[0], D[0])       # RMSE between first matrices
        0.3430603410873006
        >>> rmse_matrices(C, D)             # RMSE between each of the four matrices
        array([0.34306034, 0.25662067, 0.31842239, 0.48274156])
        >>> rmse_matrices(C, D, element_wise=True)  # RMSE element-wise along first dimension
        array([[0.22022923, 0.3886001 ],
               [0.46130561, 0.2407136 ],
               [0.38114819, 0.40178899]])
        >>> rmse_matrices(C[0], D[0], element_wise=True)
        0.3430603410873006
        >>> rmse_matrices(C[0], D[0])
        0.3430603410873006
    """
    A = np.copy(A)
    B = np.copy(B)
    if A.shape != B.shape:
        raise ValueError("Both arrays must have the same shape.")
    if A.ndim == 2:
        return np.sqrt(np.mean((A - B)**2))
    if element_wise:
        return np.sqrt(np.mean((A - B)**2, axis=0))
    return np.sqrt(np.mean(np.mean((A - B)**2, axis=2), axis=1))


# -*- coding: utf-8 -*-
"""
Attitude estimation via gravity acceleration measurements.

The simplest way to estimate the attitude from the gravitational acceleration
is using 3D `geometric quadrants <https://en.wikipedia.org/wiki/Quadrant_(plane_geometry)>`_.

Although some methods use ``arctan`` to estimate the angles :cite:p:`stm4509`
:cite:p:`fisher2010` it is preferred to use ``arctan2`` to explore all
quadrants searching the tilt angles.

First, we normalize the gravity vector, so that it has magnitude equal to 1.
Then, we get the angles to the main axes with `arctan2
<https://en.wikipedia.org/wiki/Atan2>`_ :cite:p:`pedley2013` :cite:p:`trimpe2012`:

.. math::
    \\begin{array}{ll}
    \\theta &= \\mathrm{arctan2}(a_y, a_z) \\\\
    \\phi &= \\mathrm{arctan2}\\big(-a_x, \\sqrt{a_y^2+a_z^2}\\big)
    \\end{array}

where :math:`\\theta` is the **roll** angle, :math:`\\phi` is the **pitch**
angle, and :math:`\\mathbf{a}=\\begin{bmatrix}a_x & a_y & a_z\\end{bmatrix}^T`
is the normalized vector of measured accelerations, which means
:math:`\\|\\mathbf{a}\\|=1`.

The attitude in terms of these two angles is called the **tilt**.

**Heading angle**

The heading angle, a.k.a. **yaw**, cannot be obtained from the measured
acceleration, and a different reference shall be used to obtain it. The most
common is the use of the geomagnetic information, in other words, `Earth's
magnetic field <https://en.wikipedia.org/wiki/Earth%27s_magnetic_field>`_.

With the pitch and roll angles estimated from the accelerometer, we can rotate
a magnetometer reading :math:`\\mathbf{m}=\\begin{bmatrix}m_x & m_y & m_z\\end{bmatrix}^T`,
and estimate the yaw angle :math:`\\psi` to update the orientation.

The vector :math:`\\mathbf{b}=\\begin{bmatrix}b_x & b_y & b_z\\end{bmatrix}^T`
represents the magnetometer readings after *rotating them back* to the plane,
where :math:`\\theta = \\phi = 0`.

.. math::
    \\begin{array}{cl}
    \\mathbf{b} &=
    R_y(-\\theta)R_x(-\\phi)\\mathbf{m} = R_y(\\theta)^TR_x(\\phi)^T\\mathbf{m} \\\\
    &=
    \\begin{bmatrix}
        \\cos\\theta & \\sin\\theta\\sin\\phi & \\sin\\theta\\cos\\phi \\\\
        0 & \\cos\\phi & -\\sin\\phi \\\\
        -\\sin\\theta & \\cos\\theta\\sin\\phi & \\cos\\theta\\cos\\phi
    \\end{bmatrix}
    \\begin{bmatrix}m_x \\\\ m_y \\\\ m_z\\end{bmatrix} \\\\
    \\begin{bmatrix}b_x \\\\ b_y \\\\ b_z\\end{bmatrix} &=
    \\begin{bmatrix}
        m_x\\cos\\theta + m_y\\sin\\theta\\sin\\phi + m_z\\sin\\theta\\cos\\phi \\\\
        m_y\\cos\\phi - m_z\\sin\\phi \\\\
        -m_x\\sin\\theta + m_y\\cos\\theta\\sin\\phi + m_z\\cos\\theta\\cos\\phi
    \\end{bmatrix}
    \\end{array}

Where :math:`\\mathbf{m}=\\begin{bmatrix}m_x & m_y & m_z\\end{bmatrix}^T` is
the *normalized* vector of the measured magnetic field, which means
:math:`\\|\\mathbf{m}\\|=1`.

The yaw angle :math:`\\psi` is the tilt-compensated heading angle relative to
magnetic North, computed as :cite:p:`ozyagcilar2015`:

.. math::
    \\begin{array}{ll}
    \\psi &= \\mathrm{arctan2}(-b_y, b_x) \\\\
    &= \\mathrm{arctan2}\\big(m_z\\sin\\phi - m_y\\cos\\phi, \\; m_x\\cos\\theta + \\sin\\theta(m_y\\sin\\phi + m_z\\cos\\phi)\\big)
    \\end{array}

Finally, we transform the roll-pitch-yaw angles to a quaternion representation:

.. math::
    \\mathbf{q} =
    \\begin{pmatrix}q_w\\\\q_x\\\\q_y\\\\q_z\\end{pmatrix} =
    \\begin{pmatrix}
        \\cos\\Big(\\frac{\\phi}{2}\\Big)\\cos\\Big(\\frac{\\theta}{2}\\Big)\\cos\\Big(\\frac{\\psi}{2}\\Big) + \\sin\\Big(\\frac{\\phi}{2}\\Big)\\sin\\Big(\\frac{\\theta}{2}\\Big)\\sin\\Big(\\frac{\\psi}{2}\\Big) \\\\
        \\sin\\Big(\\frac{\\phi}{2}\\Big)\\cos\\Big(\\frac{\\theta}{2}\\Big)\\cos\\Big(\\frac{\\psi}{2}\\Big) - \\cos\\Big(\\frac{\\phi}{2}\\Big)\\sin\\Big(\\frac{\\theta}{2}\\Big)\\sin\\Big(\\frac{\\psi}{2}\\Big) \\\\
        \\cos\\Big(\\frac{\\phi}{2}\\Big)\\sin\\Big(\\frac{\\theta}{2}\\Big)\\cos\\Big(\\frac{\\psi}{2}\\Big) + \\sin\\Big(\\frac{\\phi}{2}\\Big)\\cos\\Big(\\frac{\\theta}{2}\\Big)\\sin\\Big(\\frac{\\psi}{2}\\Big) \\\\
        \\cos\\Big(\\frac{\\phi}{2}\\Big)\\cos\\Big(\\frac{\\theta}{2}\\Big)\\sin\\Big(\\frac{\\psi}{2}\\Big) - \\sin\\Big(\\frac{\\phi}{2}\\Big)\\sin\\Big(\\frac{\\theta}{2}\\Big)\\cos\\Big(\\frac{\\psi}{2}\\Big)
    \\end{pmatrix}

Setting the property ``as_angles`` to ``True`` will avoid this last conversion
returning the attitude as angles.

"""

import numpy as np
from ..common.quaternion import Quaternion
from ..common.quaternion import QuaternionArray
from ..utils.core import _assert_numerical_iterable

def _assert_representation(representation: str):
    if not isinstance(representation, str):
        raise TypeError(f"Representation must be a string. Got {type(representation)}.")
    if representation not in ['quaternion', 'angles', 'rotmat']:
        raise ValueError(f"Given representation '{representation}' is NOT valid. Try 'rotmat', 'angles' or 'quaternion'")

class Tilt:
    """
    Gravity-based estimation of attitude.

    Parameters
    ----------
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of gravitational acceleration.
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of local geomagnetic field.
    representation : str, default: ``'quaternion'``
        Attitude representation. Options are ``'quaternion'``, ``'angles'`` or
        ``'rotmat'``.

    Attributes
    ----------
    acc : numpy.ndarray
        N-by-3 array with N tri-axial accelerometer samples.
    mag : numpy.ndarray
        N-by-3 array with N tri-axial magnetometer samples.
    Q : numpy.ndarray, default: None
        N-by-4 or N-by-3 array with N quaternions.

    Raises
    ------
    ValueError
        When shape of input arrays are not (3,) or (N, 3).

    Examples
    --------
    Assuming we have 3-axis accelerometer data in N-by-3 arrays, we can simply
    give these samples to the constructor. The tilt estimation works solely
    with accelerometer samples.

    >>> from ahrs.filters import Tilt
    >>> tilt = Tilt(acc_data)

    The estimated quaternions are saved in the attribute ``Q``.

    >>> tilt.Q
    array([[0.76901856, 0.60247641, -0.16815772, 0.13174072],
           [0.77310283, 0.59724644, -0.16900433, 0.1305612 ],
           [0.7735134,  0.59644005, -0.1697294,  0.1308748 ],
           ...,
           [0.7800751,  0.59908629, -0.14315079, 0.10993772],
           [0.77916118, 0.59945374, -0.14520157, 0.11171197],
           [0.77038613, 0.61061868, -0.14375869, 0.11394512]])
    >>> tilt.Q.shape
    (1000, 4)

    If we desire to estimate each sample independently, we call the
    corresponding method with each sample individually.

    >>> tilt = Tilt()
    >>> num_samples = len(acc_data)
    >>> Q = np.zeros((num_samples, 4))  # Allocate quaternions array
    >>> for t in range(num_samples):
    ...     Q[t] = tilt.estimate(acc_data[t])
    ...
    >>> tilt.Q[:5]
    array([[0.76901856, 0.60247641, -0.16815772, 0.13174072],
           [0.77310283, 0.59724644, -0.16900433, 0.1305612 ],
           [0.7735134,  0.59644005, -0.1697294,  0.1308748 ],
           [0.77294791, 0.59913005, -0.16502363, 0.12791369],
           [0.76936935, 0.60323746, -0.16540014, 0.12968487]])

    Originally, this estimation computes first the Roll-Pitch-Yaw angles and
    then converts them to Quaternions. If we desire the angles instead, we set
    it so in the parameters.

    >>> tilt = Tilt(acc_data, representation='angles')
    >>> type(tilt.Q), tilt.Q.shape
    (<class 'numpy.ndarray'>, (1000, 3))
    >>> tilt.Q[:5]
    array([[8.27467200e-04,  4.36167791e-06, 0.00000000e+00],
           [9.99352822e-04,  8.38015258e-05, 0.00000000e+00],
           [1.30423484e-03,  1.72201573e-04, 0.00000000e+00],
           [1.60337482e-03,  8.53081042e-05, 0.00000000e+00],
           [1.98459171e-03, -8.34729603e-05, 0.00000000e+00]])

    .. note::
        It will return the angles, in radians, following the standard order
        ``Roll->Pitch->Yaw``.

    The yaw angle is, expectedly, equal to zero, because the heading cannot be
    estimated with the gravity acceleration only.

    For this reason, magnetometer data can be used to estimate the yaw. This is
    also implemented and the magnetometer will be taken into account when given
    as parameter.

    >>> tilt = Tilt(acc=acc_data, mag=mag_data, representation='angles')
    >>> tilt.Q[:5]
    array([[8.27467200e-04,  4.36167791e-06, -4.54352439e-02],
           [9.99352822e-04,  8.38015258e-05, -4.52836926e-02],
           [1.30423484e-03,  1.72201573e-04, -4.49355365e-02],
           [1.60337482e-03,  8.53081042e-05, -4.44276770e-02],
           [1.98459171e-03, -8.34729603e-05, -4.36931634e-02]])

    """
    def __init__(self, acc: np.ndarray = None, mag: np.ndarray = None, **kwargs):
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.representation: str = kwargs.get('representation', 'quaternion')
        self.as_angles: bool = kwargs.get('as_angles', self.representation == 'angles') # Old parameter. Backwards compatiblity.
        _assert_representation(self.representation)
        self.angles: np.ndarray = None
        if self.acc is not None:
            self.Q = self._compute_all()

    def _compute_all(self) -> np.ndarray:
        """
        Estimate the orientation given all data.

        Attributes ``acc`` and ``mag`` must contain data. It is assumed that
        these attributes have the same shape (M, 3), where M is the number of
        observations.

        The full estimation is vectorized, to avoid the use of a time-wasting
        loop.

        Returns
        -------
        Q : numpy.ndarray
            M-by-4 array with all estimated quaternions, where M is the number
            of samples. It returns an M-by-3 array if ``representation`` is set
            to ``angles``, or M-by-3-by-3 array if set to ``rotmat``.

        """
        _assert_numerical_iterable(self.acc, 'Gravitational acceleration vector')
        self.acc = np.copy(self.acc)
        if self.acc.ndim < 2:
            return self.estimate(self.acc, self.mag, self.representation)
        # Normalization of 2D arrays
        a = self.acc/np.linalg.norm(self.acc, axis=1)[:, None]
        self.angles = np.zeros((len(a), 3))   # Allocation of angles array
        # Estimate tilt angles
        self.angles[:, 0] = np.arctan2(a[:, 1], a[:, 2])
        self.angles[:, 1] = np.arctan2(-a[:, 0], np.sqrt(a[:, 1]**2 + a[:, 2]**2))
        if self.mag is not None:
            _assert_numerical_iterable(self.mag, 'Geomagnetic field vector')
            # Estimate heading angle
            m = self.mag/np.linalg.norm(self.mag, axis=1)[:, None]
            my2 = m[:, 2]*np.sin(self.angles[:, 0]) - m[:, 1]*np.cos(self.angles[:, 0])
            mz2 = m[:, 1]*np.sin(self.angles[:, 0]) + m[:, 2]*np.cos(self.angles[:, 0])
            mx3 = m[:, 0]*np.cos(self.angles[:, 1]) + mz2*np.sin(self.angles[:, 1])
            self.angles[:, 2] = np.arctan2(my2, mx3)
        # Return angles in radians
        if self.representation == 'angles':
            return self.angles
        # RPY to Quaternion
        cp = np.cos(0.5*self.angles[:, 1])
        sp = np.sin(0.5*self.angles[:, 1])
        cr = np.cos(0.5*self.angles[:, 0])
        sr = np.sin(0.5*self.angles[:, 0])
        cy = np.cos(0.5*self.angles[:, 2])
        sy = np.sin(0.5*self.angles[:, 2])
        Q = np.zeros((len(self.acc), 4))
        Q[:, 0] = cy*cp*cr + sy*sp*sr
        Q[:, 1] = cy*cp*sr - sy*sp*cr
        Q[:, 2] = sy*cp*sr + cy*sp*cr
        Q[:, 3] = sy*cp*cr - cy*sp*sr
        if self.representation == 'quaternion':
            return Q
        Q = QuaternionArray(Q)
        return Q.to_DCM()

    def estimate(self, acc: np.ndarray, mag: np.ndarray = None, representation: str = 'quaternion') -> np.ndarray:
        """
        Estimate the quaternion from the tilting read by an orthogonal
        tri-axial array of accelerometers.

        The orientation of the roll and pitch angles is estimated using the
        measurements of the accelerometers, and finally converted to a
        quaternion representation according to :cite:p:`Wiki_Conversions`.

        Parameters
        ----------
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2.
        mag : numpy.ndarray, default: None
            N-by-3 array with measurements of magnetic field in mT.

        Returns
        -------
        q : numpy.ndarray
            Estimated attitude.

        Examples
        --------
        >>> acc_data = np.array([4.098297, 8.663757, 2.1355896])
        >>> mag_data = np.array([-28.71550512, -25.92743566, 4.75683931])
        >>> from ahrs.filters import Tilt
        >>> tilt = Tilt()
        >>> tilt.estimate(acc=acc_data, mag=mag_data)   # Estimate attitude as quaternion
        array([0.09867706 0.33683592 0.52706394 0.77395607])

        Optionally, the attitude can be retrieved as roll-pitch-yaw angles, in
        radians.

        >>> tilt = Tilt(as_angles=True)
        >>> tilt.estimate(acc=acc_data, mag=mag_data)
        array([ 76.15281566 -24.66891862 146.02634429])

        """
        _assert_representation(representation)
        _assert_numerical_iterable(acc, 'Gravitational acceleration vector')
        acc = np.copy(acc)
        a_norm = np.linalg.norm(acc)
        if a_norm == 0:
            raise ValueError("Gravitational acceleration must be non-zero")
        ax, ay, az = acc/a_norm
        ### Tilt from Accelerometer
        ex = np.arctan2( ay, az)                        # Roll
        ey = np.arctan2(-ax, np.sqrt(ay**2 + az**2))    # Pitch
        ez = 0.0                                        # Yaw
        if mag is not None:
            _assert_numerical_iterable(mag, 'Geomagnetic field vector')
            if np.linalg.norm(mag) == 0:
                raise ValueError("Magnetic field must be non-zero")
            mx, my, mz = mag/np.linalg.norm(mag)
            # Get tilted reference frame
            by = my*np.cos(ex) - mz*np.sin(ex)
            bx = mx*np.cos(ey) + np.sin(ey)*(my*np.sin(ex) + mz*np.cos(ex))
            ez = np.arctan2(-by, bx)
        if representation == 'angles':
            return np.array([ex, ey, ez])
        #### Euler to Quaternion
        cp = np.cos(0.5*ey)
        sp = np.sin(0.5*ey)
        cr = np.cos(0.5*ex)
        sr = np.sin(0.5*ex)
        cy = np.cos(0.5*ez)
        sy = np.sin(0.5*ez)
        q = np.zeros(4)
        q[0] = cy*cp*cr + sy*sp*sr
        q[1] = cy*cp*sr - sy*sp*cr
        q[2] = sy*cp*sr + cy*sp*cr
        q[3] = sy*cp*cr - cy*sp*sr
        if representation == 'quaternion':
            return q
        q = Quaternion(q)
        return q.to_DCM()



#!/usr/bin/env python3
import os
import unittest
import numpy as np
import ahrs

class TestWMM(unittest.TestCase):
    """
    Test Magnetic Field estimation with provided values.

    WMM 2015 uses a CSV test file with values split with semicolons, whereas
    the WMM 2020 uses a TXT file with values split with spaces. The position of
    their values is different. The following table shows their differences:

    =====  =================  =================  =================
    Index  CSV File (WM2015)  TXT File (WM2020)  TXT File (WM2025)
    =====  =================  =================  =================
    0      date               date               date
    1      height (km)        height (km)        height (km)
    2      latitude (deg)     latitude (deg)     latitude (deg)
    3      longitude (deg)    longitude (deg)    longitude (deg)
    4      X (nT)             D (deg)            X (nT)
    5      Y (nT)             I (deg)            Y (nT)
    6      Z (nT)             H (nT)             Z (nT)
    7      H (nT)             X (nT)             H (nT)
    8      F (nT)             Y (nT)             F (nT)
    9      I (deg)            Z (nT)             Inclination (deg)
    10     D (deg)            F (nT)             Declination (deg)
    11     GV (deg)           dD/dt (deg/year)   GV (deg)
    12     Xdot (nT/yr)       dI/dt (deg/year)   Xdot (nT/yr)
    13     Ydot (nT/yr)       dH/dt (nT/year)    Ydot (nT/yr)
    14     Zdot (nT/yr)       dX/dt (nT/year)    Zdot (nT/yr)
    15     Hdot (nT/yr)       dY/dt (nT/year)    Hdot (nT/yr)
    16     Fdot (nT/yr)       dZ/dt (nT/year)    Fdot (nT/yr)
    17     dI/dt (deg/year)   dF/dt (nT/year)    Idot (deg/year)
    18     dD/dt (deg/year)   -/-                Ddot (deg/year)
    =====  =================  =================  =================

    """
    def _load_test_values(self, filename: str) -> np.ndarray:
        """
        Load test values from file.

        Parameters
        ----------
        filename : str
            Path to file with test values.

        Returns
        -------
        data : dict
            Dictionary with the test values.
        """
        if not filename.endswith(('.csv', '.txt')):
            raise TypeError("File type is not supported. Try a csv or txt file.")
        keys = ["date", "height", "latitude", "longitude", "X", "Y", "Z", "H",
                "F", "I", "D", "GV", "dX", "dY", "dZ", "dH", "dF", "dI", "dD"]
        if filename.endswith('.csv'):
            data = np.genfromtxt(filename, delimiter=';', skip_header=1)
            if data.shape[1] < 19:
                raise ValueError(f"File '{filename}' has incomplete data.")
        elif filename.endswith('.txt'):
            data = np.genfromtxt(filename, skip_header=1, comments='#')
            if data.shape[1] < 18:
                raise ValueError(f"File '{filename}' has incomplete data.")
            if '2020' in filename:
                keys = ["date", "height", "latitude", "longitude", "D", "I",
                        "H", "X", "Y", "Z", "F", "dD", "dI", "dH", "dX", "dY",
                        "dZ", "dF"]
        return dict(zip(keys, data.T))

    def setUp(self):
        self.wmm = ahrs.utils.WMM()
        self.base_path = os.path.dirname(os.path.abspath(__file__))

    def tearDown(self) -> None:
        del self.wmm

    def test_wmm2015(self):
        test_values = self._load_test_values(os.path.join(self.base_path, "../ahrs/utils/WMM2015/WMM2015_test_values.csv"))
        num_tests = len(test_values['date'])
        for i in range(num_tests):
            self.wmm.magnetic_field(test_values['latitude'][i], test_values['longitude'][i], test_values['height'][i], date=test_values['date'][i])
            self.assertAlmostEqual(test_values['X'][i], self.wmm.X, 1, f"Expected {test_values['X'][i]:.1f}, result {self.wmm.X:.1f}")
            self.assertAlmostEqual(test_values['Y'][i], self.wmm.Y, 1, f"Expected {test_values['Y'][i]:.1f}, result {self.wmm.Y:.1f}")
            self.assertAlmostEqual(test_values['Z'][i], self.wmm.Z, 1, f"Expected {test_values['Z'][i]:.1f}, result {self.wmm.Z:.1f}")
            self.assertAlmostEqual(test_values['I'][i], self.wmm.I, 2, f"Expected {test_values['I'][i]:.2f}, result {self.wmm.I:.2f}")
            self.assertAlmostEqual(test_values['D'][i], self.wmm.D, 2, f"Expected {test_values['D'][i]:.2f}, result {self.wmm.D:.2f}")
            self.assertAlmostEqual(test_values['GV'][i], self.wmm.GV, 2, f"Expected {test_values['GV'][i]:.2f}, result {self.wmm.GV:.2f}")

    def test_wmm2020(self):
        test_values = self._load_test_values(os.path.join(self.base_path, "../ahrs/utils/WMM2020/WMM2020_TEST_VALUES.txt"))
        num_tests = len(test_values['date'])
        for i in range(num_tests):
            self.wmm.magnetic_field(test_values['latitude'][i], test_values['longitude'][i], test_values['height'][i], date=test_values['date'][i])
            self.assertAlmostEqual(test_values['X'][i], self.wmm.X, 1, f"Expected {test_values['X'][i]:.1f}, result {self.wmm.X:.1f}")
            self.assertAlmostEqual(test_values['Y'][i], self.wmm.Y, 1, f"Expected {test_values['Y'][i]:.1f}, result {self.wmm.Y:.1f}")
            self.assertAlmostEqual(test_values['Z'][i], self.wmm.Z, 1, f"Expected {test_values['Z'][i]:.1f}, result {self.wmm.Z:.1f}")
            self.assertAlmostEqual(test_values['I'][i], self.wmm.I, 2, f"Expected {test_values['I'][i]:.2f}, result {self.wmm.I:.2f}")
            self.assertAlmostEqual(test_values['D'][i], self.wmm.D, 2, f"Expected {test_values['D'][i]:.2f}, result {self.wmm.D:.2f}")

    def test_wmm2025(self):
        test_values = self._load_test_values(os.path.join(self.base_path, "../ahrs/utils/WMM2025/WMM2025_TEST_VALUES.txt"))
        num_tests = len(test_values['date'])
        for i in range(num_tests):
            self.wmm.magnetic_field(test_values['latitude'][i], test_values['longitude'][i], test_values['height'][i], date=test_values['date'][i])
            self.assertAlmostEqual(test_values['X'][i], self.wmm.X, 1, f"Expected {test_values['X'][i]:.1f}, result {self.wmm.X:.1f}")
            self.assertAlmostEqual(test_values['Y'][i], self.wmm.Y, 1, f"Expected {test_values['Y'][i]:.1f}, result {self.wmm.Y:.1f}")
            self.assertAlmostEqual(test_values['Z'][i], self.wmm.Z, 1, f"Expected {test_values['Z'][i]:.1f}, result {self.wmm.Z:.1f}")
            self.assertAlmostEqual(test_values['I'][i], self.wmm.I, 2, f"Expected {test_values['I'][i]:.2f}, result {self.wmm.I:.2f}")
            self.assertAlmostEqual(test_values['D'][i], self.wmm.D, 2, f"Expected {test_values['D'][i]:.2f}, result {self.wmm.D:.2f}")

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This novel estimator proposed by :cite:p:`wu2018-2`, offers an extremely
simplified computation of `Davenport's <../davenport.html>`_ solution to
`Wahba's problem <https://en.wikipedia.org/wiki/Wahba%27s_problem>`_, where the
full solution is reduced to a couple of floating point operations, without
losing much accuracy, and sparing computing time.

The accelerometer and magnetometer have their normalized **measurements**
:math:`^b\\mathbf{a}=\\begin{bmatrix}a_x&a_y&a_z\\end{bmatrix}^T`,
:math:`^b\\mathbf{m}=\\begin{bmatrix}m_x&m_y&m_z\\end{bmatrix}^T` in the body
frame :math:`b`.

Their corresponding normalized **reference** vectors
:math:`^r\\mathbf{a}=\\begin{bmatrix}0&0&1\\end{bmatrix}^T` and
:math:`^r\\mathbf{m}=\\begin{bmatrix}m_N&0&m_D\\end{bmatrix}^T` are such that:

.. math::
    a_x^2+a_y^2+a_z^2 = m_x^2+m_y^2+m_z^2 = m_N^2+m_D^2 = 1

We define a rotation matrix :math:`\\mathbf{C}\\in SO(3)` that transforms the
reference vectors into the body frame:

.. math::

    \\mathbf{C}\\,^r\\mathbf{a}=\\,^b\\mathbf{a}

and

.. math::

    \\mathbf{C}\\,^r\\mathbf{m}=\\,^b\\mathbf{m}

Our goal is to find the rotation matrix :math:`\\mathbf{C}` that minimizes the
error between the measured and reference vectors

.. math::

    \\begin{array}{rcl}
    \\|\\,^b\\mathbf{a}-\\mathbf{C}\\,^r\\mathbf{a}\\|^2 & \\simeq & 0 \\\\ \\\\
    \\|\\,^b\\mathbf{m}-\\mathbf{C}\\,^r\\mathbf{m}\\|^2 & \\simeq & 0
    \\end{array}

which is equivalent to solving the following optimization problem:

.. math::
    \\mathrm{min} \\big(w\\|\\,^b\\mathbf{a}-\\mathbf{C}\\,^r\\mathbf{a}\\|^2+(1-w)\\|\\,^b\\mathbf{m}-\\mathbf{C}\\,^r\\mathbf{m}\\|^2\\big)

where :math:`w` is the weight of the accelerometer correlation and :math:`1-w`
is the one of the magnetometer.

The solution to Wahba's problem is equivalent to finding the eigenvector of the
maximum eigenvalue of Davenport's matrix :math:`\\mathbf{K}`:

.. math::
    \\mathbf{K} =
    \\begin{bmatrix}
        \\mathbf{B}+\\mathbf{B}-\\mathrm{tr}(\\mathbf{B})\\mathbf{I}_3 & \\mathbf{z} \\\\ && \\\\
        \\mathbf{z}^T & \\mathrm{tr}(\\mathbf{B})
    \\end{bmatrix}

where :math:`\\mathrm{tr}` is the `matrix trace <https://en.wikipedia.org/wiki/Trace_(linear_algebra)>`_,
:math:`\\mathbf{I}_3` is the :math:`3\\times 3` `identity matrix <https://en.wikipedia.org/wiki/Identity_matrix>`_,
and the helper arrays are:

.. math::
    \\begin{array}{rcl}
    \\mathbf{B} &=& w\\,^b\\mathbf{a}\\,^r\\mathbf{a}^T + (1-w)\\,^b\\mathbf{m}\\,^r\\mathbf{m}^T \\\\
    \\mathbf{z} &=& \\begin{bmatrix}B_{23}-B_{32}\\\\B_{31}-B_{13}\\\\B_{12}-B_{21}\\end{bmatrix}
    \\end{array}

in which :math:`B_{ij}` stands for the element of :math:`\\mathbf{B}` in the
:math:`i`-th row and the :math:`j`-th column.

.. note::
    Indexing is normally starting from zero, especially in computational setups,
    but the article starts indexing from one, and it is kept like that in this
    documentation to coincide with the original document.

The eigenvalues of :math:`\\mathbf{K}` are given by:

.. math::
    \\begin{array}{rcl}
    \\lambda_{\\mathbf{K},1} &=& \\sqrt{(1-w)^2+w^2+2w(1-w)(\\alpha m_D+V)} \\\\ && \\\\
    \\lambda_{\\mathbf{K},2} &=& \\sqrt{(1-w)^2+w^2+2w(1-w)(\\alpha m_D-V)} \\\\ && \\\\
    \\lambda_{\\mathbf{K},3} &=& -\\sqrt{(1-w)^2+w^2+2w(1-w)(\\alpha m_D-V)} \\\\ && \\\\
    \\lambda_{\\mathbf{K},4} &=& -\\sqrt{(1-w)^2+w^2+2w(1-w)(\\alpha m_D+V)}
    \\end{array}

where

.. math::
    \\begin{array}{rcl}
    \\alpha &=& a_xm_x + a_ym_y + a_zm_z \\\\
    V &=& m_N\\sqrt{1-\\alpha^2}
    \\end{array}

The local `geomagnetic dip angle  <https://en.wikipedia.org/wiki/Magnetic_dip>`_
:math:`\\theta\\in[-\\frac{\\pi}{2}, \\frac{\\pi}{2}]` ensures that
:math:`m_N=\\cos\\theta>0` and :math:`\\lambda_{\\mathbf{K},1}>\\lambda_{\\mathbf{K},2}>\\lambda_{\\mathbf{K},3}>\\lambda_{\\mathbf{K},4}`.

Therefore, the attitude quaternion should be the eigenvector associated to the
eigenvalue :math:`\\lambda_{\\mathbf{K},1}`.

The dip angle is not required in the accelerometer-magnetometer configuration,
since :math:`m_D=\\alpha` and :math:`m_N=\\sqrt{1-\\alpha^2}` always holds, and
the fundamental solution to :math:`(\\mathbf{K}-\\mathbf{I})\\mathbf{q}=0` is:

.. math::
    \\begin{array}{rcl}
    q_w &=& -a_y(m_N+m_x) + a_xm_y \\\\ && \\\\
    q_x &=& (a_z-1)(m_N+m_x) + a_x(m_D-m_z) \\\\ && \\\\
    q_y &=& (a_z-1)m_y + a_y(m_D-m_z) \\\\ && \\\\
    q_z &=& a_zm_D - a_xm_N - m_z
    \\end{array}

which shows that the weights are not even necessary. Finally, the normalized
quaternion representing the attitude is:

.. math::
    \\mathbf{q} = \\frac{1}{\\sqrt{q_w^2+q_x^2+q_y^2+q_z^2}}\\begin{pmatrix}q_w & q_x & q_y & q_z\\end{pmatrix}

This estimator is extremely short and relies solely on linear operations,
making it very suitable for low-cost and simple processors. Its accuracy is
comparable to that of QUEST and FQA, but it is one order of magnitude faster.

"""

import numpy as np
from ..common.quaternion import Quaternion
from ..common.quaternion import QuaternionArray

from ..utils.core import _assert_acc_mag_inputs

class SAAM:
    """
    Super-fast Attitude from Accelerometer and Magnetometer

    Parameters
    ----------
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of acceleration in in m/s^2
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of magnetic field in nT
    representation : str, default: ``'quaternion'``
        Attitude representation. Options are ``'rotmat'`` or ``'quaternion'``.

    Attributes
    ----------
    acc : numpy.ndarray
        N-by-3 array with N accelerometer samples.
    mag : numpy.ndarray
        N-by-3 array with N magnetometer samples.
    Q : numpy.ndarray, default: None
        N-by-4 Array with all estimated orientations as quaternions, where N is
        the number of samples. Equal to None when no estimation is performed.
    A : numpy.ndarray, default: None
        3-by-3 or N-by-3-by-3 Array with all estimated orientarions as rotation
        matrices, where N is the number of rotations.

    Raises
    ------
    ValueError
        When dimension of input arrays ``acc`` and ``mag`` are not equal.

    Examples
    --------
    >>> acc_data.shape, mag_data.shape      # NumPy arrays with sensor data
    ((1000, 3), (1000, 3))
    >>> from ahrs.filters import SAAM
    >>> orientation = SAAM(acc=acc_data, mag=mag_data)
    >>> orientation.Q.shape                 # Estimated attitudes as Quaternions
    (1000, 4)
    >>> orientation.Q
    array([[-0.09867706, -0.33683592, -0.52706394, -0.77395607],
           [-0.10247491, -0.33710813, -0.52117549, -0.77732433],
           [-0.10082646, -0.33658091, -0.52082828, -0.77800078],
           ...,
           [-0.78760687, -0.57789515,  0.2131519,  -0.01669966],
           [-0.78683706, -0.57879487,  0.21313092, -0.02142776],
           [-0.77869223, -0.58616905,  0.22344478, -0.01080235]])

    """
    def __init__(self, acc: np.ndarray = None, mag: np.ndarray = None, representation='quaternion'):
        self._guard_clauses_parameters(representation)
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.representation: str = representation
        self.Q: np.ndarray = None
        if self.acc is not None and self.mag is not None:
            self.Q = self._compute_all(self.acc, self.mag)
            if representation == 'rotmat':
                self.A = Quaternion(self.Q).to_DCM() if self.Q.ndim < 2 else QuaternionArray(self.Q).to_DCM()

    def _guard_clauses_parameters(self, representation: str) -> None:
        if not isinstance(representation, str):
            raise TypeError(f"Representation must be a string. Got {type(representation)}.")
        if representation.lower() not in ['rotmat', 'quaternion']:
            raise ValueError(f"Given representation '{representation}' is NOT valid. Try 'quaternion', or 'rotmat'")

    def _compute_all(self, acc: np.ndarray, mag: np.ndarray) -> np.ndarray:
        """
        Estimate the quaternions given all data.

        Attributes ``acc`` and ``mag`` must contain data. It is assumed that
        these attributes have the same shape (M, 3), where M is the number of
        observations.

        The full estimation is vectorized, to avoid the use of a time-wasting
        loop.

        Parameters
        ----------
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer.
        mag : numpy.ndarray
            Sample of tri-axial Magnetometer.

        Returns
        -------
        Q : numpy.ndarray
            M-by-4 Array with all estimated quaternions, where M is the number
            of samples.

        """
        _assert_acc_mag_inputs(self.acc, self.mag)
        acc, mag = np.copy(self.acc), np.copy(self.mag)
        if acc.ndim < 2:
            return self.estimate(acc, mag)
        # Normalize measurements (eq. 1)
        ax, ay, az = np.transpose(acc/np.linalg.norm(acc, axis=1)[:, None])
        mx, my, mz = np.transpose(mag/np.linalg.norm(mag, axis=1)[:, None])
        # Dynamic magnetometer reference vector (eq. 12)
        mD = ax*mx + ay*my + az*mz
        mN = np.sqrt(1-mD**2)
        # Quaternion components (eq. 16)
        qw = ax*my - ay*(mN+mx)
        qx = (az-1)*(mN+mx) + ax*(mD-mz)
        qy = (az-1)*my + ay*(mD-mz)
        qz = az*mD - ax*mN-mz
        # Final quaternion (eq. 18)
        Q = np.c_[-qw, qx, qy, qz]
        return Q/np.linalg.norm(Q, axis=1)[:, None]

    def estimate(self, acc: np.ndarray, mag: np.ndarray) -> np.ndarray:
        """
        Attitude Estimation

        Parameters
        ----------
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer.
        mag : numpy.ndarray
            Sample of tri-axial Magnetometer.

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion.

        Examples
        --------
        >>> acc_data = np.array([4.098297, 8.663757, 2.1355896])
        >>> mag_data = np.array([-28.71550512, -25.92743566, 4.75683931])
        >>> from ahrs.filters import SAAM
        >>> saam = SAAM()
        >>> saam.estimate(acc=acc_data, mag=mag_data)   # Estimate attitude as quaternion
        array([-0.09867706, -0.33683592, -0.52706394, -0.77395607])

        """
        # Normalize measurements (eq. 1)
        a_norm = np.linalg.norm(acc)
        m_norm = np.linalg.norm(mag)
        if not a_norm > 0 or not m_norm > 0:    # handle NaN
            return None
        ax, ay, az = acc/a_norm
        mx, my, mz = mag/m_norm
        # Dynamic magnetometer reference vector (eq. 12)
        mD = ax*mx + ay*my + az*mz
        mN = np.sqrt(1-mD**2)
        # Quaternion components (eq. 16)
        qw = ax*my - ay*(mN+mx)
        qx = (az-1)*(mN+mx) + ax*(mD-mz)
        qy = (az-1)*my + ay*(mD-mz)
        qz = az*mD - ax*mN-mz
        # Final quaternion (eq. 18)
        q = np.array([-qw, qx, qy, qz])
        return q / np.linalg.norm(q)



# -*- coding: utf-8 -*-
"""
In 1965 `Grace Wahba <https://en.wikipedia.org/wiki/Grace_Wahba>`_ came up with
a simple, yet very intuitive, way to describe the problem of finding a rotation
between two coordinate systems.

Given a set of :math:`N` vector measurements :math:`\\mathbf{u}` in the body
coordinate system, an optimal attitude matrix :math:`\\mathbf{A}` would
minimize the `loss function <https://en.wikipedia.org/wiki/Loss_function>`_:

.. math::
    L(\\mathbf{A}) = \\frac{1}{2}\\sum_{i=1}^Nw_i|u_i-\\mathbf{A}v_i|^2

where :math:`u_i` is the i-th vector measurement in the body frame, :math:`v_i`
is the i-th vector in the reference frame, and :math:`w_i` are a set of :math:`N`
nonnegative weights for each observation. This famous formulation is known as
`Wahba's problem <https://en.wikipedia.org/wiki/Wahba%27s_problem>`_.

A first elegant solution was proposed by :cite:p:`davenport1968` that solves
this in terms of quaternions, yielding a unique optimal solution. The
corresponding **objective function** is defined as:

.. math::
    g(\\mathbf{A}) = 1 - L(\\mathbf{A}) = \\sum_{i=1}^Nw_i\\mathbf{U}^T\\mathbf{AV}

The objective function is at maximum when the loss function :math:`L(\\mathbf{A})`
is at minimum. The goal is, then, to find the optimal attitude matrix
:math:`\\mathbf{A}`, which *maximizes* :math:`g(\\mathbf{A})`. We first notice
that:

.. math::
    \\begin{array}{rl}
    g(\\mathbf{A}) =& \\sum_{i=1}^Nw_i\\mathrm{tr}\\big(\\mathbf{U}_i^T\\mathbf{AV}_i\\big) \\\\
    =& \\mathrm{tr}(\\mathbf{AB}^T)
    \\end{array}

where :math:`\\mathrm{tr}` denotes the `trace
<https://en.wikipedia.org/wiki/Trace_(linear_algebra)>`_ of a matrix, and
:math:`\\mathbf{B}` is the *attitude profile matrix*:

.. math::
    \\mathbf{B} = \\sum_{i=1}^Nw_i\\mathbf{UV}

Now, we must parametrize the attitude matrix in terms of a quaternion
:math:`\\mathbf{q}` :cite:p:`lerner1978` :

.. math::
    \\mathbf{A}(\\mathbf{q}) = (q_w^2-\\mathbf{q}_v\\cdot\\mathbf{q}_v)\\mathbf{I}_3+2\\mathbf{q}_v\\mathbf{q}_v^T-2q_w\\lfloor\\mathbf{q}\\rfloor_\\times

where :math:`\\mathbf{I}_3` is a :math:`3\\times 3` identity matrix, and the
expression :math:`\\lfloor \\mathbf{x}\\rfloor_\\times` is the `skew-symmetric
matrix <https://en.wikipedia.org/wiki/Skew-symmetric_matrix>`_ of a vector
:math:`\\mathbf{x}`. See the `quaternion page <../quaternion.html>`_ for further
details about this representation mapping.

The objective function, in terms of quaternion, becomes:

.. math::
    g(\\mathbf{q}) = (q_w^2-\\mathbf{q}_v\\cdot\\mathbf{q}_v)\\mathrm{tr}\\mathbf{B}^T + 2\\mathrm{tr}\\big(\\mathbf{q}_v\\mathbf{q}_v^T\\mathbf{B}^T\\big) + 2q_w\\mathrm{tr}(\\lfloor\\mathbf{q}\\rfloor_\\times\\mathbf{B}^T)

A simpler expression, using helper quantities, can be a bilinear relationship
of the form:

.. math::
    g(\\mathbf{q}) = \\mathbf{q}^T\\mathbf{Kq}

where the :math:`4\\times 4` matrix :math:`\\mathbf{K}` is built with:

.. math::
    \\mathbf{K} = \\begin{bmatrix}
    \\sigma & \\mathbf{z}^T \\\\
    \\mathbf{z} & \\mathbf{S}-\\sigma\\mathbf{I}_3
    \\end{bmatrix}

using the intermediate values:

.. math::
    \\begin{array}{rcl}
    \\sigma &=& \\mathrm{tr}\\mathbf{B} \\\\
    \\mathbf{S} &=& \\mathbf{B}+\\mathbf{B}^T \\\\
    \\mathbf{z} &=& \\begin{bmatrix}B_{23}-B_{32} \\\\ B_{31}-B_{13} \\\\ B_{12}-B_{21}\\end{bmatrix}
    \\end{array}

The optimal quaternion :math:`\\hat{\\mathbf{q}}`, which parametrizes the
optimal attitude matrix, is an eigenvector of :math:`\\mathbf{K}`. With the
help of `Lagrange multipliers <https://en.wikipedia.org/wiki/Lagrange_multiplier>`_,
:math:`g(\\mathbf{q})` is maximized if the eigenvector corresponding to the
largest eigenvalue :math:`\\lambda` is chosen.

.. math::
    \\mathbf{K}\\hat{\\mathbf{q}} = \\lambda\\hat{\\mathbf{q}}

The biggest disadvantage of this method is its computational load in the last
step of computing the eigenvalues and eigenvectors to find the optimal
quaternion.

"""

import numpy as np
from ..common.constants import MUNICH_LONGITUDE
from ..common.constants import MUNICH_LATITUDE
from ..common.constants import MUNICH_HEIGHT
from ..common.mathfuncs import cosd
from ..common.mathfuncs import sind
from ..utils.core import _assert_numerical_iterable
from ..utils.core import _assert_acc_mag_inputs
from ..utils.wmm import WMM
from ..utils.wgs84 import WGS

# Reference Observations in Munich, Germany
GRAVITY = WGS().normal_gravity(MUNICH_LATITUDE, MUNICH_HEIGHT)
wmm = WMM(latitude=MUNICH_LATITUDE, longitude=MUNICH_LONGITUDE, height=MUNICH_HEIGHT)
REFERENCE_MAGNETIC_VECTOR = np.array([wmm.X, wmm.Y, wmm.Z])

class Davenport:
    """
    Davenport's q-Method for attitude estimation

    Parameters
    ----------
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of acceleration in in m/s^2
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of magnetic field in mT
    weights : array-like
        Array with two weights used in each observation.
    magnetic_dip : float
        Magnetic Inclination angle, in degrees. Defaults to magnetic dip of
        Munich, Germany.
    gravity : float
        Normal gravity, in m/s^2. Defaults to normal gravity in Munich,
        Germany.

    Attributes
    ----------
    acc : numpy.ndarray
        N-by-3 array with N accelerometer samples.
    mag : numpy.ndarray
        N-by-3 array with N magnetometer samples.
    w : numpy.ndarray
        Weights of each observation.

    Raises
    ------
    ValueError
        When dimension of input arrays ``acc`` and ``mag`` are not equal.

    """
    def __init__(self, acc: np.ndarray = None, mag: np.ndarray = None, **kw) -> None:
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.w: np.ndarray = kw.get('weights', np.ones(2))
        # Reference measurements
        mdip: np.ndarray = kw.get('magnetic_dip')           # Magnetic dip, in degrees
        self.m_q: np.ndarray = REFERENCE_MAGNETIC_VECTOR if mdip is None else np.array([cosd(mdip), 0., sind(mdip)])
        g: float = kw.get('gravity', GRAVITY)               # Earth's normal gravity, in m/s^2
        self.g_q: np.ndarray = np.array([0.0, 0.0, g])      # Normal Gravity vector
        if self.acc is not None and self.mag is not None:
            self.Q: np.ndarray = self._compute_all()

    def _compute_all(self) -> np.ndarray:
        """
        Estimate all quaternions given all data.

        Attributes ``acc`` and ``mag`` must contain data.

        Returns
        -------
        Q : array
            M-by-4 Array with all estimated quaternions, where M is the number
            of samples.

        """
        _assert_acc_mag_inputs(self.acc, self.mag)
        self.acc = np.copy(self.acc)
        self.mag = np.copy(self.mag)
        if self.acc.ndim < 2:
            return self.estimate(self.acc, self.mag)
        num_samples = len(self.acc)
        return np.array([self.estimate(self.acc[t], self.mag[t]) for t in range(num_samples)])

    def estimate(self, acc: np.ndarray = None, mag: np.ndarray = None) -> np.ndarray:
        """
        Attitude Estimation

        Parameters
        ----------
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2
        mag : numpy.ndarray
            Sample of tri-axial Magnetometer in T

        Returns
        -------
        q : numpy.ndarray
            Estimated attitude as a quaternion.

        """
        _assert_numerical_iterable(acc, 'Gravitational acceleration vector')
        _assert_numerical_iterable(mag, 'Geomagnetic field vector')
        B = self.w[0]*np.outer(acc, self.g_q) + self.w[1]*np.outer(mag, self.m_q)   # Attitude profile matrix
        sigma = B.trace()
        z = np.array([B[1, 2]-B[2, 1], B[2, 0]-B[0, 2], B[0, 1]-B[1, 0]])
        S = B+B.T
        K = np.zeros((4, 4))
        K[0, 0] = sigma
        K[1:, 1:] = S - sigma*np.eye(3)
        K[0, 1:] = K[1:, 0] = z
        w, v = np.linalg.eig(K)
        return v[:, np.argmax(w)]       # Eigenvector associated to largest eigenvalue is optimal quaternion



#!/usr/bin/env python3
import unittest
import numpy as np
import ahrs

# Often used constants
SQ22 = np.sqrt(2)/2

class TestGeometry(unittest.TestCase):
    def test_geometry_circle_default(self):
        circle = ahrs.common.geometry.circle([0, 0])
        self.assertEqual(circle.max(), 1.0)
        self.assertEqual(circle.min(), -1.0)
        self.assertEqual(np.ptp(circle), 2.0)
        self.assertEqual(circle.shape, (21, 2))
        self.assertAlmostEqual(circle[:-1, 0].mean(), 0.0)
        self.assertAlmostEqual(circle[:-1, 1].mean(), 0.0)

    def test_geometry_circle_custom(self):
        circle = ahrs.common.geometry.circle([1, 1], 2.0, 10)
        self.assertEqual(circle.max(), 3.0)
        self.assertEqual(circle.min(), -1.0)
        self.assertEqual(np.ptp(circle), 4.0)
        self.assertEqual(circle.shape, (11, 2))
        self.assertAlmostEqual(circle[:-1, 0].mean(), 1.0)
        self.assertAlmostEqual(circle[:-1, 1].mean(), 1.0)

    def test_geometry_ellipse_default(self):
        ellipse = ahrs.common.geometry.ellipse([0, 0], 0.0, [1.0, 0.5])
        self.assertEqual(ellipse.max(), 1.0)
        self.assertEqual(ellipse.min(), -1.0)
        self.assertEqual(np.ptp(ellipse), 2.0)
        self.assertEqual(ellipse.shape, (21, 2))
        self.assertAlmostEqual(ellipse[:-1, 0].mean(), 0.0)
        self.assertAlmostEqual(ellipse[:-1, 1].mean(), 0.0)

    def test_geometry_ellipse_custom(self):
        ellipse = ahrs.common.geometry.ellipse([1, 1], 0.0, [2.0, 1.0], 10)
        self.assertEqual(ellipse.max(), 3.0)
        self.assertEqual(ellipse.min(), -1.0)
        self.assertEqual(np.ptp(ellipse), 4.0)
        self.assertEqual(ellipse.shape, (11, 2))
        self.assertAlmostEqual(ellipse[:-1, 0].mean(), 1.0)
        self.assertAlmostEqual(ellipse[:-1, 1].mean(), 1.0)

class TestFrames(unittest.TestCase):
    def setUp(self):
        self.lla_coords = [48.8562, 2.3508, 67.4]
        self.ecef_coords = [4201000, 172460, 4780100]
        self.ecef_decimal_tol = 1   # 1 decimal place = tolerance of 10 cm

    def test_geodetic2ecef(self):
        np.testing.assert_array_almost_equal(ahrs.common.frames.geodetic2ecef(0.0, 0.0, 0.0), [ahrs.EARTH_EQUATOR_RADIUS, 0.0, 0.0])
        np.testing.assert_array_almost_equal(ahrs.common.frames.geodetic2ecef(90.0, 0.0, 0.0), [0.0, 0.0, ahrs.EARTH_POLAR_RADIUS], decimal=4)
        np.testing.assert_array_almost_equal(ahrs.common.frames.geodetic2ecef(57.02929569, 9.950248114, 56.95), [3426949.397, 601195.852, 5327723.994], decimal=3)  # Example from GNU Octave mapping package

    def test_geodetic2ecef_wrong_inputs(self):
        self.assertRaises(ValueError, ahrs.common.frames.geodetic2ecef, 91.0, 0.0, 0.0)
        self.assertRaises(ValueError, ahrs.common.frames.geodetic2ecef, 0.0, 181.0, 0.0)

    def test_ecef2geodetic(self):
        np.testing.assert_array_almost_equal(ahrs.common.frames.ecef2geodetic(*self.ecef_coords), self.lla_coords, decimal=self.ecef_decimal_tol)

    def test_ecef2lla(self):
        np.testing.assert_array_almost_equal(ahrs.common.frames.ecef2lla(*self.ecef_coords), self.lla_coords, decimal=self.ecef_decimal_tol)

    def test_ecef2enu(self):
        np.testing.assert_array_almost_equal(ahrs.common.frames.ecef2enu(660930.192761082, -4701424.222957011, 4246579.604632881, 42, -82, 200),
                                             [186.27752, 286.84222, 939.69262],
                                             decimal=4)
        np.testing.assert_array_almost_equal(ahrs.common.frames.ecef2enu(5507528.9, 4556224.1, 6012820.8, 45.9132, 36.7484, 1877753.2),
                                             [355601.2616, -923083.1558, 1041016.4238],
                                             decimal=4)

    def test_enu2ecef(self):
        np.testing.assert_array_almost_equal(ahrs.common.frames.enu2ecef(186.27752, 286.84222, 939.69262, 42, -82, 200),
                                            [660930.192761082, -4701424.222957011, 4246579.604632881],
                                             decimal=4)
        np.testing.assert_array_almost_equal(ahrs.common.frames.enu2ecef(355601.2616, -923083.1558, 1041016.4238, 45.9132, 36.7484, 1877753.2),
                                             [5507528.9, 4556224.1, 6012820.8],
                                             decimal=4)

    def test_geodetic2enu(self):
        np.testing.assert_array_almost_equal(ahrs.common.frames.geodetic2enu(45.976, 7.658, 4531, 46.017, 7.750, 1673),
                                             [-7134.8, -4556.3, 2852.4],
                                             decimal=1)

    def test_aer2enu(self):
        np.testing.assert_array_almost_equal(ahrs.common.frames.aer2enu(34.1160, 4.1931, 15.1070),
                                             [8.4504, 12.4737, 1.1046],
                                             decimal=4)

    def test_enu2aer(self):
        np.testing.assert_array_almost_equal(ahrs.common.frames.enu2aer(8.4504, 12.4737, 1.1046),
                                             [34.1160, 4.1931, 15.1070],
                                             decimal=4)

class TestMathFuncs(unittest.TestCase):
    def test_sind(self):
        self.assertEqual(ahrs.common.mathfuncs.sind(0.0), 0.0)
        self.assertEqual(ahrs.common.mathfuncs.sind(90.0), 1.0)
        self.assertAlmostEqual(ahrs.common.mathfuncs.sind(-120.0), -0.86602540378)
        np.testing.assert_array_almost_equal(ahrs.common.mathfuncs.sind([0, 45, 90, 180, 270, 360]), [0.0, SQ22, 1.0, 0.0, -1.0, 0.0])

    def test_cosd(self):
        self.assertEqual(ahrs.common.mathfuncs.cosd(0.0), 1.0)
        self.assertAlmostEqual(ahrs.common.mathfuncs.cosd(90.0), 0.0)
        self.assertAlmostEqual(ahrs.common.mathfuncs.cosd(-120.0), -0.5)
        np.testing.assert_array_almost_equal(ahrs.common.mathfuncs.cosd([0, 45, 90, 180, 270, 360]), [1.0, SQ22, 0.0, -1.0, 0.0, 1.0])

    def test_skew(self):
        a = [1, 2, 3]
        skew = ahrs.common.mathfuncs.skew(a)
        self.assertEqual(skew.shape, (3, 3))
        np.testing.assert_allclose(skew, [[0., -3., 2.], [3., 0., -1.], [-2., 1., 0.]])

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Geometrical functions
---------------------

References
----------
.. [W1] Wikipedia: https://de.wikipedia.org/wiki/Ellipse#Ellipsengleichung_(Parameterform)
.. [WAE] Wolfram Alpha: Ellipse. (http://mathworld.wolfram.com/Ellipse.html)

"""

import numpy as np
from typing import Union

def circle(center: Union[list, np.ndarray], radius: float = 1.0, num_points: int = 20) -> np.ndarray:
    """
    Build a circle with the given characteristics.

    Parameters
    ----------
    c : array-like
        2D Coordinates of center.
    r : float
        Radius of the circle.
    num_points : int
        Number of points to build.

    Returns
    -------
    points : numpy.ndarray
        N-by-2 array with the coordinates of the circle.

    """
    R = np.linspace(0.0, 2.0*np.pi, num_points+1)
    x = center[0] + radius*np.cos(R)
    y = center[1] + radius*np.sin(R)
    return np.array([x, y]).transpose()

def ellipse(center: Union[list, np.ndarray], phi: float, axes: Union[list, np.ndarray], num_points: int = 20) -> np.ndarray:
    """
    Build an ellipse with the given characteristics.

    Parameters
    ----------
    center : array-like
        2D Coordinates of center.
    phi : float
        Angle, in radians, of the major axis w.r.t. the X-axis
    axes : array-like
        Lengths of major and minor axes, respectively.
    num_points : int
        Number of points. Defaults to 20.

    Returns
    -------
    points : numpy.ndarray
        N-by-2 array with the coordinates of the ellipse.

    """
    R = np.linspace(0.0, 2.0*np.pi, num_points+1)
    a, b = axes
    x = center[0] + a*np.cos(R)*np.cos(phi) - b*np.sin(R)*np.sin(phi)
    y = center[1] + a*np.cos(R)*np.sin(phi) + b*np.sin(R)*np.cos(phi)
    return np.array([x, y]).transpose()



"""
The main utility of the World Magnetic Model (WMM) :cite:p:`chulliat2020` is to
provide magnetic declination for any desired location on the globe.

In addition to the magnetic declination, the WMM also provides the complete
geometry of the field from 1 km below the World Geodetic System (WGS 84)
:cite:p:`wgs84-2014` ellipsoid surface to approximately 850 km above it. The
magnetic field extends deep into the Earth and far out into space, but the WMM
is not valid at these extremes.

Earth's magnetic field is viewed as a magnetic dipole produced by a sphere of
uniform magnetization. The "south" of the dipole lies on the northern
Hemipshere and drifts westward at about 55 to 60 km per year, whereas the other
pole, just outside the Antarctic Circle, drifts by about 10 to 15 km per year.

The strongest contribution to Earth's magnetism is the magnetic field produced
by the Earth's liquid-iron outer core, called the "core field". Magnetic
minerals in the crust and upper mantle make a further contribution that can be
locally significant. All these fields of "internal" origin and their large
scale components are included in the WMM.

"External" magnetic fields, arising from electric currents in the upper
atmosphere and near-Earth space, predominantly generated by solar wind, are
time-varying and produce secondary magnetic fields, which are not represented
in the WMM.

Current estimations include secondary thermal convection currents near Earth's
core-mantle producing local magnetic dipoles. These dipoles are superimposed to
approximate the observed multipole nature of the magnetic field. The creation
and decay of these inner convections are suspected to yield the magnetic drift.

The magnetic flux density (magnetic field) :math:`B` can be expressed as the
gradient of a potential, :math:`V`:

.. math::
    B = -\\nabla V

This has a convenient solution expressed in spherical harmonics of :math:`n`
degrees:

.. math::
    V(r, \\theta, \\phi) = a \\sum_{n=1}^k\\Big(\\frac{a}{r}\\Big)^{n+1}\\sum_{m=0}^{n}\\big(g_n^m \\cos(m\\phi) + h_n^m \\sin(m\\phi)\\big) P_n^m(\\theta)

where :math:`a` is Earth's mean radius; :math:`g_n^m` and :math:`h_n^m` are
*Gaussian coefficients* of degree :math:`n` and order :math:`m`; :math:`r`,
:math:`\\theta`, and :math:`\\phi` are the geocentric radius, coelevation and
longitude in spherical polar coordinates; and :math:`P_n^m(\\theta)` are the
associated Legendre functions :cite:p`hofmann2006`.

The time-dependent Gauss coefficients are estimated empirically from a
least-squares fit using satellite magnetic measurements :cite:p:`langel1998`.
These coefficients are provided by the NCEI Geomagnetic Modeling Team and
British Geological Survey in a file with extension COF :cite:p:`chulliat2020`.

With degree :math:`n=1` only dipoles are considered. For :math:`n=2` the
quadrupoles, :math:`n=3` the octuploles, and so on. The method of this WMM
expands to *degree* and *order* 12.

The secular variation (SV) is the yearly change of the core field, which is
also accounted in the WMM by a linear model. Due to unpredictable changes in
the core field, the values of the WMM coefficients are updated every five
years (a lustrum). The most recent version is valid from 2020 to 2029.

The geomagnetic field vector B is described by 7 elements:

+---------+-----------------------------------------------+--------+----------------+
|         |                                               |        | Range          |
| Element | Definition                                    | Units  +--------+-------+
|         |                                               |        | Min    | Max   |
+=========+===============================================+========+========+=======+
| X       | Northerly intensity                           | nT     | -17000 | 43000 |
+---------+-----------------------------------------------+--------+--------+-------+
| Y       | Easterly intensity                            | nT     | -18000 | 17000 |
+---------+-----------------------------------------------+--------+--------+-------+
| Z       | Vertical intensity (Positive downwards)       | nT     | -67000 | 62000 |
+---------+-----------------------------------------------+--------+--------+-------+
| H       | Horizontal intensity                          | nT     |      0 | 43000 |
+---------+-----------------------------------------------+--------+--------+-------+
| F       | Total intensity                               | nT     |  23000 | 67000 |
+---------+-----------------------------------------------+--------+--------+-------+
| I       | Inclination angle (a.k.a. dip angle)          | degree |    -90 | 90    |
+---------+-----------------------------------------------+--------+--------+-------+
| D       | Declination angle (a.k.a. magnetic variation) | degree |   -180 | 180   |
+---------+-----------------------------------------------+--------+--------+-------+

The quantities ``X``, ``Y`` and ``Z`` are perpendicular vectors and can be used
to determine the quantities ``F``, ``I`` and ``D``, and viceversa.

The vertical direction is perpendicular to the horizontal plane of the WGS 84
ellipsoid model of the Earth.

At a location on the plane of a chosen horizontal coordinate system,
**grivation** is the angle between grid north and magnetic north, i.e., the
angle measured clockwise from the direction parallel to the grid's Northing
axis to the horizontal component of the magnetic field at the observer's
location.

Grivation is useful for local surveys, where location is given by grid
coordinates rather than by longitude and latitude. It is dependent on the map
projection used to define the grid coordinates. In general, it is estimated as:

.. math::
    GV = D - C

where :math:`C` is the "convergence-of-meridians" defined as the clockwise
angle from the northward meridional arc to the grid Northing direction.

The class ``WMM`` contains a couple of methods to load and create a geomagnetic
model from a set of given coefficients. To obtain the magnetic elements at a
certain spot on the Earth, we call the method ``magnetic_field``

.. code:: python

    >>> wmm = ahrs.utils.WMM()              # Create today's magnetic model
    >>> wmm.magnetic_field(10.0, -20.0)     # Magnetic field at latitude = 10°, longitude = -20°
    >>> wmm.D                               # Magnetic declination [degrees]
    -9.122361367239034
    >>> wmm.magnetic_field(10.0, -20.0, height=10.5)     # 10.5 km above sea level
    >>> wmm.D
    -9.128404039098971

By default, the class ``WMM`` will create a model for the day, when the object
is being created. To ask for the values at a different date, we simply pass it
as a parameter.

.. code:: python

    >>> wmm.magnetic_field(10.0, -20.0, height=10.5, date=datetime.date(2017, 5, 12))    # on 12th May, 2017
    >>> wmm.D
    -9.73078560629778

All main elements are computed at the same time and accessed independently

.. code:: python

    >>> wmm.X
    30499.640469609083
    >>> wmm.Y
    -5230.267158472566
    >>> wmm.Z
    -1716.633311360368
    >>> wmm.H
    30944.850352270452
    >>> wmm.F
    30992.427998627096
    >>> wmm.I
    -3.1751692563622993
    >>> wmm.GV
    -9.73078560629778

or in a dictionary

.. code:: python

    >>> wmm.magnetic_elements
    {'X': 30499.640469609083, 'Y': -5230.267158472566, 'Z': -1716.633311360368,
    'H': 30944.850352270452, 'F': 30992.427998627096, 'I': -3.1751692563622993,
    'D': -9.73078560629778, 'GV': -9.73078560629778}

.. note::
    The model in this package includes coefficients for dates between **2015**
    and **2029** only. Models out of this timespan cannot be built.

The WMM was developed jointly by the National Centers for Environmental
Information (NCEI, Boulder CO, USA) (formerly National Geophysical Data Center
(NGDC)) and the British Geological Survey (BGS, Edinburgh, Scotland). As part
of the regular update cycle of the World Magnetic Model both institutions have
released the latest model on December 10th, 2019.

This script is based on the originally conceived one by Christopher Weiss
(cmweiss@gmail.com) :cite:p:`weiss2016`, who adapted it from the geomagc
software and World Magnetic Model of the NOAA Satellite and Information
Service, National Geophysical Data Center :cite:p:`chulliat2020`.

License
-------
The WMM source code and binaries are in the public domain and not licensed or
under copyright. The information and software may be used freely by the public.
As required by 17 U.S.C. 403, third parties producing copyrighted works
consisting predominantly of the material produced by U.S. government agencies
must provide notice with such work(s) identifying the U.S. Government material
incorporated and stating that such material is not subject to copyright
protection.

"""

# PSL
import datetime
import pkgutil
from io import StringIO
from typing import Union, Tuple, Dict

# Third-Party Dependencies
import numpy as np
from ..common.constants import EARTH_EQUATOR_RADIUS
from ..common.constants import EARTH_POLAR_RADIUS
from ..common.constants import EARTH_MEAN_RADIUS
from ..common.constants import MUNICH_LATITUDE
from ..common.constants import MUNICH_LONGITUDE
from ..common.constants import MUNICH_HEIGHT
from ..common.constants import DEG2RAD
from ..common.constants import RAD2DEG
from ..common.frames import ned2enu

def geodetic2spherical(lat: float, lon: float, h: float, a: float = EARTH_EQUATOR_RADIUS/1000.0, b: float = EARTH_POLAR_RADIUS/1000.0) -> Tuple[float, float, float]:
    """
    Transform geodetic coordinates into spherical geocentric coordinates

    The transformation cannot be a simple cylindric to spherical conversion, as
    we must also consider a planet's ellipsoid form. With the aid of its
    pre-defined flatness and eccentricity, we can better approximate the values
    of the conversion.

    In this function the Earth's major and minor semi-axis are considered.
    However, we can convert the coordinates of different ellipsoidal bodies, by
    giving the dimensions of its semi-axes.

    Notice that the longitude between both systems remains the same.

    Parameters
    ----------
    lat : float
        Latitude, in radians, of point in geodetic coordinates
    lon : float
        Longitude, in radians, of point in geodetic coordinates
    h : float
        Height, in kilometers, of point in geodetic coordinates
    a : float, default: 6378.137
        Major semi-axis, in kilometers. Defaults to Earth's equatorial radius
    b : float, default: 6356.752314245
        Minor semi-axis, in kilometers. Defaults to Earth's polar radius

    Returns
    -------
    lat_spheric : float
        Latitude of point in spherical coordinates.
    lon : float
        Longitue of point in spherical coordinates. Same as geodetic.
    r : float
        Radial distance of point in spherical coordinates.
    """
    # Estimate Spheroid's Flatness and First Eccentricity
    f = (a-b)/a                             # Flatness
    e2 = f*(2.0-f)                          # First Eccentricity
    # Transform geodetic coordinates into spherical geocentric coordinates
    Rc = a/np.sqrt(1.0-e2*np.sin(lat)**2)   # Radius of curvature of prime vertical
    rho = (Rc+h)*np.cos(lat)
    z = (Rc*(1-e2)+h)*np.sin(lat)
    r = np.linalg.norm([rho, z])            # Radial distance
    lat_spheric = np.arcsin(z/r)            # Spherical latitude
    return lat_spheric, lon, r

class WMM:
    """
    This class computes all elements of the World Magnetic Model (WMM) at any
    given point on Earth.

    The main magnetic field :math:`B` is a potential field defined, in
    geocentric spherical coordinates (longitude :math:`\\lambda`, latitude
    :math:`\\phi '` and radius :math:`r`), as the negative spatial gradient of a
    scalar potential at a time :math:`t`. This potential can be expanded in
    terms of spherical harmonics:

    .. math::
        V(\\lambda, \\phi', r, t) = a\\sum_{n=1}^{N}\\Big(\\frac{a}{r}\\Big)^{n+1}\\sum_{m=0}^{n}f(n, m, \\lambda, t)P_n^m(\\phi')

    where

    .. math::
        f(n, m, \\lambda, t) = g_n^m(t) \\cos(m\\lambda) + h_n^m(t) \\sin(m\\lambda)

    and the Schmidt semi-normalized associated Legendre functions :math:`P_n^m(\\phi')`
    are defined as:

    .. math::
        P_n^m(\\mu) = \\left\\{
        \\begin{array}{ll}
            \\sqrt{2\\frac{(n-m)!}{(n+m)!}}P_{n, m}(\\mu) & \\mathrm{if} \\; m > 0 \\\\
            P_{n, m}(\\mu) & \\mathrm{if} \\; m = 0
        \\end{array}
        \\right.

    Any object of this class is initialized with the corresponding epoch,
    determined by the given date. If no date is given, it is assumed for the
    day of the object's creation.

    Once the WMM object is created, the estimation of the geomagnetic elements
    is carried out with a call to the method `magnetic_field` giving the
    location on Earth at which the magnetic elements will be calculated. This
    location is given in decimal geodetic coordinates. See examples.

    Every WMM object is created with a set of coefficients read from a COF
    file, defined by the desired working date of the model. The latest
    model available is WMM2025 corresponding to the lustrum 2025-2029.

    This class can create models with dates between 2015 and 2029.

    Parameters
    ----------
    date : datetime.date, int or float, default: current day
        Date of desired magnetic field estimation.
    latitude : float, default: 48.137230
        Latitude, in decimal degrees, in geodetic coordinates.
    longitude : float, default: 11.575508
        Longitude, in decimal degrees, in geodetic coordinates.
    height : float, default: 0.521
        Mean Sea Level Height, in kilometers.
    frame : str, default: 'NED'
        Local tangent plane coordinate frame. Valid options are right-handed
        ``'NED'`` for North-East-Down and ``'ENU'`` for East-North-Up.

    Attributes
    ----------
    date : datetime.date, default: datetime.date.today()
        Desired date to estimate
    date_dec : float
        Desired date to estimate as decimal
    epoch : float
        Initial time of model in decimal years
    model : str
        WMM Model identificator
    modeldate : str
        Release date of WMM Model
    wmm_filename : str
        COF File used to build Model
    degree : int
        Degree of model
    latitude : float
        Latitude, in decimal degrees, in geodetic coordinates
    longitude : float
        Longitude in decimal degrees, in geodetic coordinates
    height : float, default: 0.0
        Mean Sea Level Height in kilometers
    X : float, default: None
        Northerly intensity, in nT
    Y : float, default: None
        Easterly intensity, in nT
    Z : float, default: None
        Vertical intensity, in nT
    H : float, default: None
        Horizontal intensity, in nT
    F : float, default: None
        Total intensity, in nT
    I : float, default: None
        Inclination angle (dip), in degrees
    D : float, default: None
        Declination angle, in degrees
    GV : float, default: None
        Grivation, in degrees

    Examples
    --------
    The magnetic field can be computed at the creation of the WMM object by
    passing the main parameters to its constructor:

    >>> wmm = ahrs.utils.WMM(datetime.date(2017, 5, 12), latitude=10.0, longitude=-20.0, height=10.5)
    >>> wmm.magnetic_elements
    {'X': 30499.640469609083, 'Y': -5230.267158472566, 'Z': -1716.633311360368,
    'H': 30944.850352270452, 'F': 30992.427998627096, 'I': -3.1751692563622993,
    'D': -9.73078560629778, 'GV': -9.73078560629778}

    """
    def __init__(self,
                 date: Union[datetime.date, int, float] = None,
                 latitude: float = MUNICH_LATITUDE,
                 longitude: float = MUNICH_LONGITUDE,
                 height: float = MUNICH_HEIGHT/1000,
                 frame: str = 'NED') -> None:
        self.epoch: float = 2015
        self.reset_coefficients(date)
        self.__dict__.update(dict.fromkeys(['X', 'Y', 'Z', 'H', 'F', 'I', 'D', 'GV']))
        self.latitude: float = latitude
        self.longitude: float = longitude
        self.height: float = height
        self.frame: str = frame
        self._guard_clauses()
        if all([self.latitude, self.longitude]):
            self.magnetic_field(self.latitude, self.longitude, self.height, date=self.date)

    def _guard_clauses(self) -> None:
        for item in ['latitude', 'longitude', 'height']:
            if not isinstance(self.__getattribute__(item), (int, float)):
                raise TypeError(f"{item} must be int or float. Got {type(self.__getattribute__(item))}")
        if not isinstance(self.frame, str):
            raise TypeError(f"frame must be str. Got {type(self.frame)}")
        if abs(self.latitude) > 90:
            raise ValueError(f"Latitude must be between -90 and 90. Got {self.latitude}")
        if abs(self.longitude) > 180:
            raise ValueError(f"Longitude must be between -180 and 180. Got {self.longitude}")
        if self.frame.upper() not in ['NED', 'ENU']:
            raise ValueError(f"{self.frame} is not a valid frame. Try 'NED' or 'ENU'")

    def reset_coefficients(self, date: Union[datetime.date, int, float] = None) -> None:
        """
        Reset Gauss coefficients to given date.

        Given the date, the corresponding coefficients are updated. Basic
        properties (epoch, release date, and model id) are read and updated in
        the current instance.

        The two coefficient tables (arrays) are also updated, where the
        attribute `c` contains the Gaussian coefficients, while the attribute
        `cd` contains the secular Gaussian coefficients.

        The lenght of the Gaussian coefficient array determines the degree
        :math:`n` of the model. This property updates the value of attribute
        ``degree``.

        Parameters
        ----------
        date : datetime.date, int or float, default: current day
            Date of desired magnetic field estimation.

        """
        self.reset_date(date)
        self.__dict__.update(self.get_properties(self.wmm_filename))
        self.load_coefficients(self.wmm_filename)

    def load_coefficients(self, cof_file: str) -> None:
        """
        Load model coefficients from COF file.

        The model coefficients, also referred to as Gauss coefficients, are
        listed in a COF file. These coefficients can be used to compute values
        of the fields elements and their annual rates of change at any
        location near the surface of the Earth.

        The COF file has 6 columns:

        * ``n`` is the degree.
        * ``m`` is the order.
        * ``g`` are time-dependent Gauss coefficients of degree ``n`` and order ``m``.
        * ``h`` are time-dependent Gauss coefficients of degree ``n`` and order ``m``.
        * ``gd`` are secular variations of coefficient ``g``.
        * ``hd`` are secular variations of coefficient ``h``.

        which constitute the *model* of the field. The first-order time
        derivatives are called *secular terms*. The units are ``nT`` for the
        main field, and ``nT/year`` for the secular variation.

        The Gauss coefficients are defined for a time :math:`t` as:

        .. math::
            \\begin{eqnarray}
            g_n^m(t) & = & g_n^m(t_0) + (t-t_0) \\dot{g}_n^m(t_0) \\\\
            h_n^m(t) & = & h_n^m(t_0) + (t-t_0) \\dot{h}_n^m(t_0)
            \\end{eqnarray}

        where time is given in decimal years and :math:`t_0` corresponds to the
        epoch read from the corresponding COF file.

        Parameters
        ----------
        cof_file : str
            Path to COF file with the coefficients of the WMM

        """
        file_data = pkgutil.get_data(__name__, cof_file).decode()
        data = np.genfromtxt(StringIO(file_data), comments="999999", skip_header=1)
        self.degree = int(max(data[:, 0]))
        self.c = np.zeros((self.degree+1, self.degree+1))
        self.cd = np.zeros((self.degree+1, self.degree+1))
        for row in data:
            n, m = row[:2].astype(int)
            self.c[m, n] = row[2]           # g_n^m
            self.cd[m, n] = row[4]          # g_n^m secular
            if m != 0:
                self.c[n, m-1] = row[3]     # h_n^m
                self.cd[n, m-1] = row[5]    # h_n^m secular

    def get_properties(self, cof_file: str) -> Dict[str, Union[str, float]]:
        """
        Return dictionary of WMM properties from COF file.

        Three properties are read and returned in a dictionary:

        * ``epoch`` is the initial time :math:`t_0` as a `float`.
        * ``model`` is a string of model used for the required lustrum.
        * ``modeldate`` is the release date of used magnetic model.

        Parameters
        ----------
        cof_file : str
            Path to COF file with the coefficients of the WMM

        Returns
        -------
        properties : dictionary
            Dictionary with the three WMM properties.

        Examples
        --------
        >>> wmm = ahrs.WMM()
        >>> wmm.get_properties('my_coefficients.COF')
        {'model': 'WMM-2020', 'modeldate': '12/10/2019', 'epoch': 2020.0}

        """
        if not cof_file.endswith(".COF"):
            raise TypeError("File must have extension 'COF'")
        first_line = pkgutil.get_data(__name__, cof_file).decode().split('\n')[0]
        v = first_line.strip().split()
        properties = dict(zip(["model", "modeldate"], v[1:]))
        properties.update({"epoch": float(v[0])})
        return properties

    def reset_date(self, date: Union[datetime.date, int, float]) -> None:
        """
        Set date to use with the model.

        The WMM requires a date. This date can be given as an instance of
        `datetime.date` or as a decimalized date of the format ``YYYY.d``.

        If None is given it sets the date to the present day. In addition, the
        corresponding COF file is also set.

        Please note that only coefficents from year 2015 and later are provided
        with this module.

        Parameters
        ----------
        date : datetime.date, int or float, default: current day
            Date of desired magnetic field estimation.
        """
        if date is None:
            self.date = datetime.date.today()
            self.date_dec = self.date.year + self.date.timetuple().tm_yday/365.0
        elif isinstance(date, (int, float)):
            self.date_dec = float(date)
            self.date = datetime.date.fromordinal(round(datetime.date(int(date), 1, 1).toordinal() + (self.date_dec-int(self.date_dec))*365))
        elif isinstance(date, datetime.date):
            self.date = date
            self.date_dec = self.date.year + self.date.timetuple().tm_yday/365.0
        else:
            raise TypeError(f"Date must be an instance of datetime.date or a decimalized year. Got {type(date)}")
        if self.date.year < 2015:
            raise ValueError("No available coefficients for dates before 2015.")
        if self.date_dec < 2020.0:
            self.wmm_filename = 'WMM2015/WMM.COF'
            return
        if self.date_dec < 2025.0:
            self.wmm_filename = 'WMM2020/WMM.COF'
            return
        self.wmm_filename = 'WMM2025/WMM.COF'

    def denormalize_coefficients(self, latitude: float) -> None:
        """
        Recursively estimate associated Legendre polynomials and derivatives
        done in a recursive way as described by Michael Plett in
        :cite:p:`plett1978` for an efficient computation.

        Given the Gaussian coefficients, it is possible to estimate the
        magnetic field at any latitude on Earth for a certain date.

        First, it is assumed that :math:`P_n^m(x)` are the Schmidt
        semi-normalized functions. A normalization is made so that the
        relative strength of terms of same degree :math:`n` but order :math:`m`
        are used by comparing their respective Gauss coefficients.

        For :math:`m=0` they are called *Legendre Polynomials* and can be
        computed recursively with:

        .. math::

            P_n(x) = \\frac{2n-1}{n} x P_{n-1}(x) - \\frac{n-1}{n}P_{n-2}(x)

        For :math:`m>0` they are known as *associated Legendre functions*
        of degree :math:`n` and order :math:`m` and reduced to:

        .. math::

            P_{nm}(x) = (1-t^2)^{m/2} \\frac{d^m P_n(x)}{dt^m}

        expressing the associated Legendre functions in terms of the Legendre
        polynomials of same degree :math:`n`.

        A more general formula to estimate both polynomial and associated
        functions is given by:

        .. math::

            P_{nm}(x) = 2^{-n}(1-x^2)^{m/2} \\sum_{k=0}^{K}(-1)^k\\frac{(2n-2k)!}{k!(n-k)!(n-m-2k)!}x^{n-m-2k}

        where :math:`K` is either :math:`(n-m)/2` or :math:`(n-m-1)/2`,
        whichever is an integer. For a computational improvement, the terms are
        calculated recursively.

        We have to denormalize the coefficients from Schmidt to Gauss. The
        Gauss functions :math:`P^{n, m}` are related to Schmidt functions
        :math:`P_n^m` as:

        .. math::

            P_n^m = S_{n, m}P^{n, m}

        where the factors :math:`S_{n, m}` are combined with Gaussian
        coefficients to accelerate the computation, because they are
        independent of the geographic location. Thus, we denormalize the
        coefficients with:

        .. math::

            \\begin{array}{ll}
                g^{n,m} & = S_{n,m} g_n^m \\\\
                h^{n,m} & = S_{n,m} h_n^m
            \\end{array}

        The recursion for :math:`S_{n, m}` is:

        .. math::

            \\begin{array}{rlr}
                S_{0,0} & = 1 & \\\\
                S_{n,0} & = S_{n-1, 0} \\frac{2n-1}{n} & n\\geq 1 \\\\
                S_{n,m} & = S_{n-1, m}\\sqrt{\\frac{(n-m+1)(\\delta _m^1+1)}{n+m}} & m\\geq 1
            \\end{array}

        where the Kronecker delta :math:`\\delta_j^i` is:

        .. math::

            \\delta_j^i = \\left\\{
            \\begin{array}{ll}
                1 & \\: i = j \\\\
                0 & \\: \\mathrm{otherwise}
            \\end{array}
            \\right.

        Similarly, :math:`P^{n, m}(x)` can be recursively obtained:

        .. math::

            \\begin{array}{ll}
                P^{0,0} & = 1 \\\\
                P^{n,n} & = \\sin (x) P^{n-1, n-1} \\\\
                P^{n,m} & = \\cos (x) P^{n-1, m} - K^{n, m} P^{n-2, m}
            \\end{array}

        where:

        .. math::

            K^{n, m} = \\left\\{
            \\begin{array}{ll}
                \\frac{(n-1)^2-m^2}{(2n-1)(2n-3)} & \\: n>1 \\\\
                0 & \\: n=1
            \\end{array}
            \\right.

        Likewise, the gradient :math:`\\frac{dP^{n, m}}{dx}` is estimated as:

        .. math::

            \\begin{array}{llr}
                \\frac{dP^{0, 0}}{dx} & = 1 & \\\\
                \\frac{dP^{n, n}}{dx} & = \\sin (x) \\frac{dP^{n-1, n-1}}{dx} + \\cos (x) P^{n-1, n-1} & n\\geq 1 \\\\
                \\frac{dP^{n, m}}{dx} & = \\cos (x) \\frac{dP^{n-1, m}}{dx} - \\sin (x) P^{n-1, m} - K^{n, m} \\frac{dP^{n-2, m}}{dx} &
            \\end{array}

        Parameters
        ----------
        latitude : float
            Latitude in spherical geocentric coordinates

        """
        cos_lat = np.cos(latitude)                  # cos(phi')
        sin_lat = np.sin(latitude)                  # sin(phi')
        S = np.identity(self.degree+1)              # Scale factors
        self.k = np.zeros((self.degree+1, self.degree+1))
        self.P = np.identity(self.degree+2)
        self.dP = np.zeros((self.degree+2, self.degree+1))
        for n in range(1, self.degree+1):
            S[0, n] = S[0, n-1] * (2*n-1)/n
            delta = 1           # Kronecker delta
            for m in range(n+1):
                self.k[m, n] = ((n-1)**2 - m**2) / ((2*n-1)*(2*n-3))
                if m > 0:
                    S[m, n] = S[m-1, n] * np.sqrt((n-m+1)*(delta+1)/(n+m))
                    self.c[n, m-1] *= S[m, n]
                    self.cd[n, m-1] *= S[m, n]
                    delta = 0
                if n == m:
                    self.P[m, n] = cos_lat*self.P[m-1, n-1]
                    self.dP[m, n] = cos_lat*self.dP[m-1, n-1] + sin_lat*self.P[m-1, n-1]
                else:
                    self.P[m, n] = sin_lat*self.P[m, n-1] - self.k[m, n]*self.P[m, n-2]
                    self.dP[m, n] = sin_lat*self.dP[m, n-1] - cos_lat*self.P[m, n-1] - self.k[m, n]*self.dP[m, n-2]
                self.c[m, n] *= S[m, n]
                self.cd[m, n] *= S[m, n]

    def magnetic_field(self, latitude: float, longitude: float, height: float = 0.0, date: Union[datetime.date, int, float] = datetime.date.today()) -> None:
        """
        Calculate the geomagnetic field elements for a location on Earth.

        The code includes comments with references to equation numbers
        corresponding to the ones in the official report.

        Having the coefficients :math:`g^{n, m}` and :math:`h^{n, m}`, we
        extrapolate them for the desired time :math:`t` as:

        .. math::

            \\begin{array}{ll}
                g_n^m(t) & = g_n^m(t_0) + \\Delta_t \\dot{g}_n^m (t_0) \\\\
                h_n^m(t) & = h_n^m(t_0) + \\Delta_t \\dot{h}_n^m (t_0)
            \\end{array}

        where :math:`\\Delta_t = t-t_0` is the difference between the time
        :math:`t` and the reference epoch model :math:`t_0` (``2020.0`` for the
        newest version.)

        The vector components of the main magnetic field :math:`B` are then
        calculated with:

        .. math::

            \\begin{array}{ll}
                X' & = -\\sum_{n=1}^N\\Big(\\frac{a}{r}\\Big)^{n+2} \\sum_{m=0}^n\\big(g_n^m(t) \\cos(m\\lambda)+h_n^m(t)\\sin(m\\lambda)\\big) \\frac{dP_n^m(\\sin \\phi ')}{d\\phi '} \\\\
                Y' & = \\frac{1}{\\cos\\phi '}\\sum_{n=1}^N\\Big(\\frac{a}{r}\\Big)^{n+2} \\sum_{m=0}^n m\\big(g_n^m(t) \\sin(m\\lambda)-h_n^m(t)\\cos(m\\lambda)\\big)P_n^m(\\sin \\phi ') \\\\
                Z' & = -\\sum_{n=1}^N(n+1)\\Big(\\frac{a}{r}\\Big)^{n+2} \\sum_{m=0}^n\\big(g_n^m(t) \\cos(m\\lambda)+h_n^m(t)\\sin(m\\lambda)\\big)P_n^m(\\sin \\phi ')
            \\end{array}

        Finally, the geomagnetic vector components are rotated into ellipsoidal
        reference frame.

        .. math::

            \\begin{array}{ll}
                X & = X'\\cos(\\phi ' - \\phi) - Z' \\sin(\\phi ' - \\phi) \\\\
                Y & = Y' \\\\
                Z & = X'\\sin(\\phi ' - \\phi) + Z' \\cos(\\phi ' - \\phi)
            \\end{array}

        These components are used to compute the rest of the magnetic elements:

        .. math::

            \\begin{array}{ll}
                H & = \\sqrt{X^2 + Y^2} \\\\
                F & = \\sqrt{H^2 + Z^2} \\\\
                I & = \\arctan(\\frac{Z}{H}) \\\\
                D & = \\arctan(\\frac{Y}{X})
            \\end{array}

        .. note::
            The use of ``arctan2`` yields a more precise result than ``arctan``,
            because it estimates the angle exploring all quadrants.

        For polar regions, where the declination changes drastically, the WMM
        defines two different grivations (one for each pole) defined as:

        .. math::

            GV = \\left\\{
            \\begin{array}{ll}
                D-\\lambda & \\: \\phi > 55 ° \\\\
                D+\\lambda & \\: \\phi < -55 °
            \\end{array}
            \\right.

        Parameters
        ----------
        latitude : float
            Latitude, in decimal degrees, in geodetic coordinates
        longitude : float
            Longitude in decimal degrees, in geodetic coordinates
        height : float, default: 0.0
            Mean Sea Level Height in kilometers
        date : datetime.date, int or float, default: datetime.date.today()
            Desired date to estimate
        """
        if date is not None:
            self.reset_coefficients(date)
        self.latitude = latitude
        self.longitude = longitude
        self.height = height
        latitude *= DEG2RAD
        longitude *= DEG2RAD
        # Transform geodetic coordinates into spherical geocentric coordinates
        lat_prime, _, r = geodetic2spherical(latitude, longitude, self.height)
        # Compute cos(m*phi') and sin(m*phi') for all m values
        self.sp = np.zeros(self.degree+1)               # sin(m*phi')
        self.cp = np.ones(self.degree+2)                # cos(m*phi')
        self.sp[1] = np.sin(longitude)
        self.cp[1] = np.cos(longitude)
        for m in range(2, self.degree+1):
            self.sp[m] = self.sp[1]*self.cp[m-1] + self.cp[1]*self.sp[m-1]
            self.cp[m] = self.cp[1]*self.cp[m-1] - self.sp[1]*self.sp[m-1]
        dt = round(self.date_dec, 1) - self.epoch       # t - t_0
        self.gh = np.zeros((self.degree+2, self.degree+1))
        self.denormalize_coefficients(lat_prime)
        cos_lat = np.cos(lat_prime)                     # cos(phi')
        sin_lat = np.sin(lat_prime)                     # sin(phi')
        Zp = Xp = Yp = Bp = 0.0
        a = EARTH_MEAN_RADIUS/1000.0                    # Mean earth radius in km
        ar = a/r
        # Spherical Harmonics (eq. 4)
        for n in range(1, self.degree+1):   #### !! According to report it must be equal to defined degree (12 not 13)
            arn2 = ar**(n+2)
            x_p = y_p = z_p = 0.0
            for m in range(n+1):            #### !! According to report it must be equal to n
                self.gh[m, n] = self.c[m, n] + dt*self.cd[m, n]     # g_n^m (eq. 9)
                # Terms of spherical harmonic expansions
                gchs = self.gh[m, n]*self.cp[m]         # g(t)cos(ml)
                gshc = self.gh[m, n]*self.sp[m]         # g(t)sin(ml)
                if m > 0:
                    self.gh[n, m-1] = self.c[n, m-1] + dt*self.cd[n, m-1]   # h_n^m (eq. 9)
                    gchs += self.gh[n, m-1]*self.sp[m]  # g(t)cos(ml) + h(t)sin(ml)
                    gshc -= self.gh[n, m-1]*self.cp[m]  # g(t)sin(ml) - h(t)cos(ml)
                x_p += gchs * self.dP[m, n]
                y_p += m * gshc * self.P[m, n]
                z_p += gchs * self.P[m, n]
                # SPECIAL CASE: NORTH/SOUTH GEOGRAPHIC POLES
                if (cos_lat == 0.0 and m == 1):
                    Bp += arn2 * gshc
                    if n > 1:
                        Bp *= sin_lat - self.k[m, n]
            Xp += arn2 * x_p                            # (eq. 10)  #### !! According to report must be a substraction. Must re-check this
            Yp += arn2 * y_p                            # (eq. 11)
            Zp -= (n+1) * arn2 * z_p                    # (eq. 12)
        Yp = Bp if cos_lat == 0.0 else Yp/cos_lat
        # Transform magnetic vector components to geodetic coordinates (eq. 17)
        self.X = Xp*np.cos(lat_prime-latitude) - Zp*np.sin(lat_prime-latitude)
        self.Y = Yp
        self.Z = Xp*np.sin(lat_prime-latitude) + Zp*np.cos(lat_prime-latitude)
        # Rotate elements if ENU frame
        if self.frame.upper() == 'ENU':
            self.X, self.Y, self.Z = ned2enu([self.X, self.Y, self.Z])
        # Total Intensity, Inclination and Declination (eq. 19)
        self.H = np.linalg.norm([self.X, self.Y])       # sqrt(X^2+Y^2)
        self.F = np.linalg.norm([self.H, self.Z])       # sqrt(H^2+Z^2)
        self.I = RAD2DEG*np.arctan2(self.Z, self.H)
        self.D = RAD2DEG*np.arctan2(self.Y, self.X)
        # Grivation (eq. 1)
        self.GV = self.D.copy()
        if self.latitude > 55.0:
            self.GV -= self.longitude
        if self.latitude < -55.0:
            self.GV += self.longitude

    @property
    def magnetic_elements(self) -> Dict[str, float]:
        """
        Main geomagnetic elements in a dictionary

        =======  =============================================
        Element  Definition
        =======  =============================================
        X        Northerly intensity
        Y        Easterly intensity
        Z        Vertical intensity (Positive downwards)
        H        Horizontal intensity
        F        Total intensity
        I        Inclination angle (a.k.a. dip angle)
        D        Declination angle (a.k.a. magnetic variation)
        GV       Grivation
        =======  =============================================

        Example
        -------
        >>> wmm = WMM(datetime.date(2017, 5, 12), latitude=10.0, longitude=-20.0, height=10.5)
        >>> wmm.magnetic_elements
        {'X': 30499.640469609083, 'Y': -5230.267158472566, 'Z': -1716.633311360368,
        'H': 30944.850352270452, 'F': 30992.427998627096, 'I': -3.1751692563622993,
        'D': -9.73078560629778, 'GV': -9.73078560629778}
        """
        return {k: self.__dict__[k] for k in ['X', 'Y', 'Z', 'H', 'F', 'I', 'D', 'GV']}

    @property
    def geodetic_vector(self) -> np.ndarray:
        """
        Magnetic field vector in geodetic coordinates

        The magnetic field vector is returned as a 3D vector in geodetic
        coordinates.

        Example
        -------
        >>> wmm = WMM(datetime.date(2017, 5, 12), latitude=10.0, longitude=-20.0, height=10.5)
        >>> wmm.geodetic_vector
        array([ 30499.64046961,  -5230.26715847,  -1716.63331136])
        """
        return np.array([self.X, self.Y, self.Z])



# -*- coding: utf-8 -*-
"""
This estimator proposed by Robert Mahony et al. :cite:p:`mahony2008` is formulated
as a deterministic kinematic observer on the Special Orthogonal group SO(3)
driven by an instantaneous attitude and angular velocity measurements
:cite:p:`hamel2006`.

By exploiting the geometry of the special orthogonal group a related observer,
the *passive complementary filter*, is derived that decouples the gyro
measurements from the reconstructed attitude in the observer inputs
:cite:p:`mahony2005`.

Direct and passive filters are extended to estimate gyro bias on-line. This
leads to an observer on SO(3), termed the **Explicit Complementary Filter**
:cite:p:`euston2008`, that requires only accelerometer and gyro outputs,
suitable for hardware implementation, and providing good estimates as well as
online gyro bias computation.

Sensor Models
-------------

The `gyroscopes <https://en.wikipedia.org/wiki/Gyroscope>`_ measure angular
velocity in the body-fixed frame, whose error
model is:

.. math::
    \\Omega_y = \\Omega + b + \\mu \\in\\mathbb{R}^3

where :math:`\\Omega` is the true angular velocity, :math:`b` is a constant (or
slow time-varying) **bias**, and :math:`\\mu` is an additive **measurement
noise**.

An ideal `accelerometer <https://en.wikipedia.org/wiki/Accelerometer>`_
measures the instantaneous linear acceleration :math:`\\dot{\\mathbf{v}}` of
the body frame *minus* the gravitational acceleration field :math:`\\mathbf{g}_0`. In
practice, the output :math:`\\mathbf{a}` of a tri-axial accelerometer has also
an added bias :math:`\\mathbf{b}_a` and noise :math:`\\mu_a`.

.. math::
    \\mathbf{a} = \\mathbf{R}^T(\\dot{\\mathbf{v}}-\\mathbf{g}_0) + \\mathbf{b}_a + \\mu_a

where :math:`|\\mathbf{g}_0|\\approx 9.8`. Under quasi-static conditions it is
common to normalize this vector so that:

.. math::
    \\mathbf{v}_a = \\frac{\\mathbf{a}}{|\\mathbf{a}|} \\approx -\\mathbf{R}^Te_3

where :math:`e_3=\\frac{\\mathbf{g}_0}{|\\mathbf{g}_0|}=\\begin{bmatrix}0 & 0 & 1\\end{bmatrix}^T`.

A `magnetometer <https://en.wikipedia.org/wiki/Magnetometer>`_ provides
measurements for the magnetic field

.. math::
    \\mathbf{m} = \\mathbf{R}^T\\mathbf{h} + \\mathbf{B}_m + \\mu_b

where :math:`\\mathbf{h}` is `Earth's magnetic field <https://en.wikipedia.org/wiki/Earth%27s_magnetic_field>`_
as measured at the inertial frame, :math:`\\mathbf{B}_m` is the local magnetic
disturbance and :math:`\\mu_b` is the measurement noise.

The magnetic intensity is irrelevant for the estimation of the attitude, and
only the **direction** of the geomagnetic field will be used. Normally, this
measurement is also normalized:

.. math::
    \\mathbf{v}_m = \\frac{\\mathbf{m}}{|\\mathbf{m}|}

The measurement vectors :math:`\\mathbf{v}_a` and :math:`\\mathbf{v}_m` are
used to build an instantaneous algebraic rotation :math:`\\mathbf{R}`:

.. math::
    \\mathbf{R} \\approx \\mathbf{R}_y=\\underset{\\mathbf{R}\\in SO(3)}{\\operatorname{arg\\,min}} (\\lambda_1|e_3-\\mathbf{Rv}_a|^2 + \\lambda_2|\\mathbf{v}_m^*-\\mathbf{Rv}_m|^2)

where :math:`\\mathbf{v}_m^*` is the referential direction of the local
magnetic field.

The corresponding weights :math:`\\lambda_1` and :math:`\\lambda_2` are chosen
depending on the relative confidence in the sensor outputs.

Two degrees of freedom in the rotation matrix are resolved using the
acceleration readings (*tilt*) and the final degree of freedom is resolved
using the magnetometer (*heading*.)

The system considered is the kinematics:

.. math::
    \\dot{\\mathbf{R}} = \\mathbf{R}\\lfloor\\Omega\\rfloor_\\times

where :math:`\\lfloor\\Omega\\rfloor_\\times` denotes the `skew-symmetric
matrix <https://en.wikipedia.org/wiki/Skew-symmetric_matrix>`_ of
:math:`\\Omega=\\begin{bmatrix}\\Omega_X & \\Omega_Y & \\Omega_Z\\end{bmatrix}^T`:

.. math::
    \\lfloor\\Omega\\rfloor_\\times = \\begin{bmatrix}
    0 & -\\Omega_Z & \\Omega_Y\\\\
    \\Omega_Z & 0 & -\\Omega_X\\\\
    -\\Omega_Y & \\Omega_X & 0
    \\end{bmatrix}

The inverse operation taking the skew-symmetric matrix into its associated
vector is :math:`\\Omega=\\mathrm{vex}(\\lfloor\\Omega\\rfloor_\\times)`.

The kinematics can also be written in terms of the quaternion representation in
SO(3):

.. math::
    \\dot{\\mathbf{q}} = \\frac{1}{2}\\mathbf{qp}(\\Omega)

where :math:`\\mathbf{q}=\\begin{pmatrix}q_w & \\mathbf{q}_v\\end{pmatrix}=\\begin{pmatrix}q_w & q_x & q_y & q_z\\end{pmatrix}`
represents a unit quaternion, and :math:`\\mathbf{p}(\\Omega)` represents the
unitary pure quaternion associated to the angular velocity
:math:`\\mathbf{p}(\\Omega)=\\begin{pmatrix}0 & \\Omega_X & \\Omega_Y & \\Omega_Z\\end{pmatrix}`

.. warning::
    The product of two quaternions :math:`\\mathbf{p}` and :math:`\\mathbf{q}`
    is the Hamilton product defined as:

    .. math::
        \\mathbf{pq} =
        \\begin{bmatrix}
        p_w q_w - p_x q_x - p_y q_y - p_z q_z \\\\
        p_w q_x + p_x q_w + p_y q_z - p_z q_y \\\\
        p_w q_y - p_x q_z + p_y q_w + p_z q_x \\\\
        p_w q_z + p_x q_y - p_y q_x + p_z q_w
        \\end{bmatrix}

Error Criteria
--------------

We denote :math:`\\hat{\\mathbf{R}}` as the *estimation* of the body-fixed
rotation matrix :math:`\\mathbf{R}`. The used estimation error is the relative
rotation from the body-fixed frame to the estimator frame:

.. math::
    \\tilde{\\mathbf{R}} := \\hat{\\mathbf{R}}^T\\mathbf{R}

Mahony's proposed observer, based on `Lyapunov stability analysis
<https://en.wikipedia.org/wiki/Lyapunov_stability>`_, yields the **cost
function**:

.. math::
    E_\\mathrm{tr} = \\frac{1}{2}\\mathrm{tr}(\\mathbf{I}_3-\\tilde{\\mathbf{R}})

The goal of attitude estimation is to provide a set of dynamics for an estimate
:math:`\\hat{\\mathbf{R}}(t)\\in SO(3)` to drive the error rotation
:math:`\\tilde{\\mathbf{R}}(t)\\to \\mathbf{I}_3`, which in turn would drive
:math:`\\hat{\\mathbf{R}}\\to\\mathbf{R}`.

The general form of the observer is termed as a **Complementary Filter on SO(3)**:

.. math::
    \\dot{\\hat{\\mathbf{R}}} = \\lfloor\\mathbf{R}\\Omega + k_P\\hat{\\mathbf{R}}\\omega\\rfloor_\\times\\hat{\\mathbf{R}}

where :math:`k_P>0` and the term :math:`\\mathbf{R}\\Omega + k_P\\hat{\\mathbf{R}}\\omega`
is expressed in the inertial frame.

The *innovation* or *correction* term :math:`\\omega`, derived from the error
:math:`\\tilde{\\mathbf{R}}`, can be thought of as a non-linear
approximation of the error between :math:`\\mathbf{R}` and :math:`\\hat{\\mathbf{R}}`
as measured from the frame of reference associated with :math:`\\hat{\\mathbf{R}}`.

When no correction term is used (:math:`\\omega=0`) the error rotation
:math:`\\tilde{\\mathbf{R}}` will be constant.

If :math:`\\tilde{\\mathbf{q}}=\\begin{pmatrix}\\tilde{q}_w & \\tilde{\\mathbf{q}}_v\\end{pmatrix}`
is the quaternion related to :math:`\\tilde{\\mathbf{R}}`, then:

.. math::
    E_\\mathrm{tr} = 2|\\tilde{\\mathbf{q}}_v|^2 = 2(1-\\tilde{q}_w^2)

Explicit Complementary Filter
-----------------------------

Let :math:`\\mathbf{v}_{0i}\\in\\mathbb{R}^3` denote a set of :math:`n\\geq 2`
known directions in the inertial (fixed) frame of reference, where the
directions are not collinear, and :math:`\\mathbf{v}_{i}\\in\\mathbb{R}^3` are
their associated measurements. The measurements are body-fixed frame
observations of the fixed inertial directions:

.. math::
    \\mathbf{v}_i = \\mathbf{R}^T\\mathbf{v}_{0i} + \\mu_i

where :math:`\\mu_i` is a process noise. We assume that :math:`|\\mathbf{v}_{0i}|=1`
and normalize all measurements to force :math:`|\\mathbf{v}_i|=1`.

For :math:`n` measures, the **global cost** becomes:

.. math::
    E_\\mathrm{mes} = \\sum_{i=1}^nk_i-\\mathrm{tr}(\\tilde{\\mathbf{R}}\\mathbf{M})

where :math:`\\mathbf{M}>0` is a positive definite matrix if :math:`n\\geq 3`,
or positive semi-definite if :math:`n=2`:

.. math::
    \\mathbf{M} = \\mathbf{R}^T\\mathbf{M}_0\\mathbf{R}

with:

.. math::
    \\mathbf{M}_0 = \\sum_{i=1}^nk_i\\mathbf{v}_{0i}\\mathbf{v}_{0i}^T

The weights :math:`k_i>0` are chosen depending on the relative confidence in
the measured directions.

Low-cost IMUs typically measure gravitational, :math:`\\mathbf{a}`, and
magnetic, :math:`\\mathbf{m}`, vector fields.

.. math::
    \\begin{array}{rcl}
    \\mathbf{v}_a &=& \\mathbf{R}^T\\frac{\\mathbf{a}_0}{|\\mathbf{a}_0|} \\\\ && \\\\
    \\mathbf{v}_m &=& \\mathbf{R}^T\\frac{\\mathbf{m}_0}{|\\mathbf{m}_0|}
    \\end{array}

In this case, the cost function :math:`E_\\mathrm{mes}` becomes:

.. math::
    E_\\mathrm{mes} = k_1(1-\\langle\\hat{\\mathbf{v}}_a, \\mathbf{v}_a\\rangle) + k_2(1-\\langle\\hat{\\mathbf{v}}_m, \\mathbf{v}_m\\rangle)

.. tip::
    When the IMU is subject to high magnitude accelerations (takeoff, landing
    manoeuvres, etc.) it may be wise to reduce the relative weighing of the
    accelerometer data (:math:`k_1 \\ll k_2`) compared to the magnetometer data.
    Conversely, when the IMU is mounted in the proximity of powerful electric
    motors leading to low confidence in the magnetometer readings choose
    :math:`k_1 \\gg k_2`.

Expressing the kinematics of the Explicit Complementary Filter as quaternions
we get:

.. math::
    \\begin{array}{rcl}
    \\dot{\\hat{\\mathbf{q}}} &=& \\frac{1}{2}\\hat{\\mathbf{q}}\\mathbf{p}\\Big(\\Omega_y-\\hat{b} + k_P\\omega_\\mathrm{mes}\\Big) \\\\ && \\\\
    \\dot{\\hat{b}} &=& -k_I\\omega_\\mathrm{mes} \\\\ && \\\\
    \\omega_\\mathrm{mes} &=& \\displaystyle\\sum_{i=1}^nk_i\\mathbf{v}_i\\times\\hat{\\mathbf{v}}_i
    \\end{array}

The estimated attitude rate of change :math:`\\dot{\\hat{\\mathbf{q}}}` is
multiplied with the sample-rate :math:`\\Delta t` to integrate the angular
displacement, which is finally added to the previous attitude, obtaining the
newest estimated attitude.

.. math::
    \\mathbf{q}_t = \\mathbf{q}_{t-1} + \\dot{\\hat{\\mathbf{q}}}_t\\Delta t

"""

import numpy as np
from ..common.quaternion import Quaternion
from ..common.orientation import acc2q
from ..common.orientation import am2q
from ..utils.core import _assert_numerical_iterable
from ..utils.core import _assert_numerical_positive_variable

class Mahony:
    """
    Mahony's Nonlinear Complementary Filter on SO(3)

    If ``acc`` and ``gyr`` are given as parameters, the orientations will be
    immediately computed with method ``updateIMU``.

    If ``acc``, ``gyr`` and ``mag`` are given as parameters, the orientations
    will be immediately computed with method ``updateMARG``.

    Parameters
    ----------
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of acceleration in m/s^2
    gyr : numpy.ndarray, default: None
        N-by-3 array with measurements of angular velocity in rad/s
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of magnetic field in mT
    frequency : float, default: 100.0
        Sampling frequency in Herz
    Dt : float, default: 0.01
        Sampling step in seconds. Inverse of sampling frequency. Not required
        if `frequency` value is given
    k_P : float, default: 1.0
        Proportional filter gain
    k_I : float, default: 0.3
        Integral filter gain
    q0 : numpy.ndarray, default: None
        Initial orientation, as a versor (normalized quaternion).

    Attributes
    ----------
    gyr : numpy.ndarray
        N-by-3 array with N gyroscope samples.
    acc : numpy.ndarray
        N-by-3 array with N accelerometer samples.
    mag : numpy.ndarray
        N-by-3 array with N magnetometer samples.
    frequency : float
        Sampling frequency in Herz.
    Dt : float
        Sampling step in seconds. Inverse of sampling frequency.
    k_P : float
        Proportional filter gain.
    k_I : float
        Integral filter gain.
    q0 : numpy.ndarray
        Initial orientation, as a versor (normalized quaternion)

    Raises
    ------
    ValueError
        When dimension of input array(s) ``acc``, ``gyr``, or ``mag`` are not
        equal.

    Examples
    --------
    Assuming we have 3-axis sensor data in N-by-3 arrays, we can simply give
    these samples to their corresponding type. The Mahony algorithm can work
    solely with gyroscope samples, although the use of accelerometer samples is
    much encouraged.

    The easiest way is to directly give the full array of samples to their
    matching parameters.

    >>> from ahrs.filters import Mahony
    >>> orientation = Mahony(gyr=gyro_data, acc=acc_data)   # Using IMU

    The estimated quaternions are saved in the attribute ``Q``.

    >>> type(orientation.Q), orientation.Q.shape
    (<class 'numpy.ndarray'>, (1000, 4))

    If we desire to estimate each sample independently, we call the
    corresponding method.

    .. code:: python

        orientation = Mahony()
        Q = np.tile([1., 0., 0., 0.], (num_samples, 1)) # Allocate for quaternions
        for t in range(1, num_samples):
            Q[t] = orientation.updateIMU(Q[t-1], gyr=gyro_data[t], acc=acc_data[t])

    Further on, we can also use magnetometer data.

    >>> orientation = Mahony(gyr=gyro_data, acc=acc_data, mag=mag_data)   # Using MARG

    This algorithm is dynamically adding the orientation, instead of estimating
    it from static observations. Thus, it requires an initial attitude to build
    on top of it. This can be set with the parameter ``q0``:

    >>> orientation = Mahony(gyr=gyro_data, acc=acc_data, q0=[0.7071, 0.0, 0.7071, 0.0])

    If no initial orientation is given, then an attitude using the first
    samples is estimated. This attitude is computed assuming the sensors are
    straped to a system in a quasi-static state.

    A constant sampling frequency equal to 100 Hz is used by default. To change
    this value we set it in its parameter ``frequency``. Here we set it, for
    example, to 150 Hz.

    >>> orientation = Mahony(gyr=gyro_data, acc=acc_data, frequency=150.0)

    Or, alternatively, setting the sampling step (:math:`\\Delta t = \\frac{1}{f}`):

    >>> orientation = Mahony(gyr=gyro_data, acc=acc_data, Dt=1/150)

    This is specially useful for situations where the sampling rate is variable:

    .. code:: python

        orientation = Mahony()
        Q = np.tile([1., 0., 0., 0.], (num_samples, 1)) # Allocate for quaternions
        for t in range(1, num_samples):
            orientation.Dt = new_sample_rate
            Q[t] = orientation.updateIMU(Q[t-1], gyr=gyro_data[t], acc=acc_data[t])

    Mahony's algorithm uses an explicit complementary filter with two gains
    :math:`k_P` and :math:`k_I` to correct the estimation of the attitude.
    These gains can be set in the parameters too:

    >>> orientation = Mahony(gyr=gyro_data, acc=acc_data, kp=0.5, ki=0.1)

    Following the experimental settings of the original article, the gains are,
    by default, :math:`k_P=1` and :math:`k_I=0.3`.

    """
    def __init__(self,
            gyr: np.ndarray = None,
            acc: np.ndarray = None,
            mag: np.ndarray = None,
            frequency: float = 100.0,
            k_P: float = 1.0,
            k_I: float = 0.3,
            q0: np.ndarray = None,
            b0: np.ndarray = None,
            **kwargs) -> None:
        self.gyr: np.ndarray = gyr
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.frequency: float = frequency
        self.q0: np.ndarray = q0
        self.b = b0 if b0 is not None else np.zeros(3)
        self.k_P: float = k_P
        self.k_I: float = k_I
        # Old parameter names for backward compatibility
        self.k_P: float = kwargs.get('kp', k_P)
        self.k_I: float = kwargs.get('ki', k_I)
        self.Dt: float = kwargs.get('Dt', 1.0/self.frequency if self.frequency else 0.01)
        self._assert_validity_of_inputs()
        # Estimate all orientations if sensor data is given
        if self.gyr is not None and self.acc is not None:
            self.Q = self._compute_all()

    def _assert_validity_of_inputs(self) -> None:
        """Asserts the validity of the inputs."""
        for item in ["frequency", "Dt", "k_P", "k_I"]:
            _assert_numerical_positive_variable(getattr(self, item), item)
        for item in ['q0', 'b']:
            if self.__getattribute__(item) is not None:
                if isinstance(self.__getattribute__(item), bool):
                    raise TypeError(f"Parameter '{item}' must be an array.")
                if not isinstance(self.__getattribute__(item), (list, tuple, np.ndarray)):
                    raise TypeError(f"Parameter '{item}' is not an array. Got {type(self.__getattribute__(item))}.")
                if np.copy(self.__getattribute__(item)).ndim != 1:
                    raise ValueError(f"Parameter '{item}' must be a 1-dimensional array.")
                self.__setattr__(item, np.copy(self.__getattribute__(item)))
        if self.q0 is not None:
            if self.q0.shape != (4,):
                raise ValueError("Parameter 'q0' must be an array with 4 elements.")
            if not np.allclose(np.linalg.norm(self.q0), 1.0):
                raise ValueError("Parameter 'q0' must be a unit quaternion.")
        if self.b is not None:
            if self.b.shape != (3,):
                raise ValueError("Parameter 'b' must be an array with 3 elements.")

    def _compute_all(self) -> np.ndarray:
        """
        Estimate the quaternions given all data

        Attributes ``gyr``, ``acc`` and, optionally, ``mag`` must contain data.

        Returns
        -------
        Q : numpy.ndarray
            M-by-4 Array with all estimated quaternions, where M is the number
            of samples.

        """
        _assert_numerical_iterable(self.gyr, 'Angular velocity vector')
        _assert_numerical_iterable(self.acc, 'Gravitational acceleration vector')
        self.gyr = np.copy(self.gyr)
        self.acc = np.copy(self.acc)
        if self.acc.shape != self.gyr.shape:
            raise ValueError("acc and gyr are not the same size")
        num_samples = len(self.gyr)
        Q = np.zeros((num_samples, 4))
        # Compute with IMU Architecture
        if self.mag is None:
            Q[0] = acc2q(self.acc[0]) if self.q0 is None else self.q0/np.linalg.norm(self.q0)
            for t in range(1, num_samples):
                Q[t] = self.updateIMU(Q[t-1], self.gyr[t], self.acc[t])
            return Q
        # Compute with MARG Architecture
        _assert_numerical_iterable(self.mag, 'Geomagnetic field vector')
        self.mag = np.copy(self.mag)
        if self.mag.shape != self.gyr.shape:
            raise ValueError("mag and gyr are not the same size")
        Q[0] = am2q(self.acc[0], self.mag[0]) if self.q0 is None else self.q0/np.linalg.norm(self.q0)
        for t in range(1, num_samples):
            Q[t] = self.updateMARG(Q[t-1], self.gyr[t], self.acc[t], self.mag[t])
        return Q

    def updateIMU(self, q: np.ndarray, gyr: np.ndarray, acc: np.ndarray, dt: float = None) -> np.ndarray:
        """
        Attitude Estimation with a IMU architecture.

        Parameters
        ----------
        q : numpy.ndarray
            A-priori quaternion.
        gyr : numpy.ndarray
            Sample of tri-axial Gyroscope in rad/s.
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2.
        dt : float, default: None
            Time step, in seconds, between consecutive Quaternions.

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion.

        Examples
        --------
        >>> orientation = Mahony()
        >>> Q = np.tile([1., 0., 0., 0.], (num_samples, 1)) # Allocate for quaternions
        >>> for t in range(1, num_samples):
        ...     Q[t] = orientation.updateIMU(Q[t-1], gyr=gyro_data[t], acc=acc_data[t])

        """
        q = Quaternion(q)
        dt = self.Dt if dt is None else dt
        if gyr is None or np.linalg.norm(gyr) == 0:
            return q.to_array()
        Omega = np.copy(gyr)
        a_norm = np.linalg.norm(acc)
        if a_norm > 0:
            R = q.to_DCM()
            v_a = R.T@np.array([0.0, 0.0, 1.0])     # Expected Earth's gravity
            # ECF
            omega_mes = np.cross(acc/a_norm, v_a)   # Cost function (eqs. 32c and 48a)
            bDot = -self.k_I*omega_mes                   # Estimated change in Gyro bias (eqs.32b and 48c)
            self.b += bDot * dt                          # Estimated Gyro bias
            Omega = Omega - self.b + self.k_P*omega_mes  # Gyro correction
        p = np.array([0.0, *Omega])
        qDot = 0.5*q.product(p)                     # Rate of change of quaternion (eqs. 45 and 48b)
        q += qDot*dt                                # Update orientation
        q /= np.linalg.norm(q)                      # Normalize Quaternion (Versor)
        return q

    def updateMARG(self, q: np.ndarray, gyr: np.ndarray, acc: np.ndarray, mag: np.ndarray, dt: float = None) -> np.ndarray:
        """
        Attitude Estimation with a MARG architecture.

        Parameters
        ----------
        q : numpy.ndarray
            A-priori quaternion.
        gyr : numpy.ndarray
            Sample of tri-axial Gyroscope in rad/s.
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2.
        mag : numpy.ndarray
            Sample of tri-axail Magnetometer in uT.
        dt : float, default: None
            Time step, in seconds, between consecutive Quaternions.

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion.

        Examples
        --------
        >>> orientation = Mahony()
        >>> Q = np.tile([1., 0., 0., 0.], (num_samples, 1)) # Allocate for quaternions
        >>> for t in range(1, num_samples):
        ...     Q[t] = orientation.updateMARG(Q[t-1], gyr=gyro_data[t], acc=acc_data[t], mag=mag_data[t])

        """
        q = Quaternion(q)
        dt = self.Dt if dt is None else dt
        if gyr is None or np.linalg.norm(gyr) == 0:
            return q.to_array()
        Omega = np.copy(gyr)
        a_norm = np.linalg.norm(acc)
        if a_norm > 0:
            m_norm = np.linalg.norm(mag)
            if m_norm == 0:
                return self.updateIMU(q, gyr, acc)
            a = np.copy(acc)/a_norm
            m = np.copy(mag)/m_norm
            R = q.to_DCM()
            v_a = R.T@np.array([0.0, 0.0, 1.0])     # Expected Earth's gravity
            # Rotate magnetic field to inertial frame
            h = R@m
            v_m = R.T@np.array([0.0, np.linalg.norm([h[0], h[1]]), h[2]])
            v_m /= np.linalg.norm(v_m)
            # ECF
            omega_mes = np.cross(a, v_a) + np.cross(m, v_m) # Cost function (eqs. 32c and 48a)
            bDot = -self.k_I*omega_mes                   # Estimated change in Gyro bias (eqs.32b and 48c)
            self.b += bDot * dt                          # Estimated Gyro bias
            Omega = Omega - self.b + self.k_P*omega_mes  # Gyro correction (eq. 48b)
        p = np.array([0.0, *Omega])
        qDot = 0.5*q.product(p)                     # Rate of change of quaternion (eqs. 45 and 48b)
        q += qDot*dt                                # Update orientation
        q /= np.linalg.norm(q)                      # Normalize Quaternion (Versor)
        return q
